from pydantic_xml import BaseXmlModel, attr, element


class TagSimple(BaseXmlModel, tag="TagSimple"):
    id: int | None = element(tag="id", default=None)


class CollectionType(BaseXmlModel):
    tag_simple: list[TagSimple] | None = element(tag="TagSimple", default=None)


class Collection(BaseXmlModel):
    set: CollectionType | None = element(tag="set", default=None)
    add: CollectionType | None = element(tag="add", default=None)
    remove: CollectionType | None = element(tag="remove", default=None)
    update: CollectionType | None = element(tag="update", default=None)


class Asset(BaseXmlModel):
    name: str | None = element(tag="name", default=None)
    tags: Collection | None = element(tag="tags", default=None)


class Data(BaseXmlModel):
    asset: Asset = element(tag="Asset")


class Criteria(BaseXmlModel):
    field: str = attr(name="field")
    operator: str = attr(name="operator")
    value: str


class Filters(BaseXmlModel):
    criteria: list[Criteria] | None = element(tag="Criteria")


class Preferences(BaseXmlModel):
    start_from_offset: int | None = element(tag="startFromOffset", default=None)
    start_from_id: int | None = element(tag="startFromId", default=None)
    limit_results: int | None = element(tag="limitResults", default=None)


class ServiceRequest(BaseXmlModel, tag="ServiceRequest"):
    data: Data | None = element(tag="data", default=None)
    filters: Filters | None = element(tag="filters", default=None)
    preferences: Preferences | None = element(tag="preferences", default=None)


def create_asset_request(
    criteria: list[Criteria] | None,
    name: str | None = None,
    add_tags: list[int] = [],
    remove_tags: list[int] = [],
    start_from_offset: int | None = None,
    start_from_id: int | None = None,
    limit_results: int | None = None,
) -> ServiceRequest:
    if add_tags or remove_tags:
        tags = Collection(
            add=CollectionType(tag_simple=[TagSimple(id=tag) for tag in add_tags]),
            remove=CollectionType(
                tag_simple=[TagSimple(id=tag) for tag in remove_tags]
            ),
        )
    else:
        tags = None
    return ServiceRequest(
        data=Data(
            asset=Asset(
                name=name,
                tags=tags,
            )
        ),
        filters=Filters(criteria=criteria),
        preferences=Preferences(
            start_from_offset=start_from_offset,
            start_from_id=start_from_id,
            limit_results=limit_results,
        ),
    )



from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class Model(BaseModel):
    model_config = ConfigDict(validate_by_name=True, alias_generator=to_camel)


class TagSimple(Model):
    id: int
    name: str


class ListItem(Model):
    tag_simple: TagSimple = Field(alias="TagSimple")


class Children(Model):
    list: list[ListItem]


class Tag(Model):
    id: int
    created: str | None = None
    modified: str | None = None
    name: str | None = None
    color: str | None = None
    children: Children | None = None


class Datum(Model):
    tag: Tag = Field(alias="Tag")


class ServiceResponse(Model):
    response_code: str
    data: list[Datum]
    count: int


class Wrapper(Model):
    service_response: ServiceResponse = Field(alias="ServiceResponse")



import datetime

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class Model(BaseModel):
    model_config = ConfigDict(validate_by_name=True, alias_generator=to_camel)


class TagSimple(Model):
    id: int
    name: str


class ListItem(Model):
    tag_simple: TagSimple = Field(alias="TagSimple")


class Tags(Model):
    list: list[ListItem]


class Asset(Model):
    id: int
    type: str | None = None
    created: datetime.datetime | None = None
    modified: datetime.datetime | None = None
    criticality_score: int | None = Field(alias="criticalityScore", default=None)
    name: str | None = None
    tags: Tags | None = None


class Datum(Model):
    asset: Asset = Field(alias="Asset")


class ServiceResponse(Model):
    data: list[Datum] | None = None
    response_code: str = Field(alias="responseCode")
    count: int | None = None


class Wrapper(Model):
    service_response: ServiceResponse = Field(alias="ServiceResponse")



from pydantic_xml import BaseXmlModel, element, wrapped, attr
import datetime
import ipaddress


class ResponseWarning(BaseXmlModel):
    code: str | None = element(tag="CODE", default=None)
    text: str = element(tag="TEXT")
    url: str | None = element(tag="URL", default=None)


class NetBIOS(BaseXmlModel):
    network_id: str = attr(name="network_id", default="0")
    value: str


class DNS(BaseXmlModel):
    network_id: str = attr(name="network_id", default="0")
    value: str


class Domain(BaseXmlModel):
    netblock: str = attr(name="netblock", default="")
    network_id: str | None = attr(name="network_id", default=None)
    value: str


class IPRange(BaseXmlModel):
    network_id: str | None = attr(name="network_id", default=None)
    value: str


class IP(BaseXmlModel):
    network_id: str | None = attr(name="network_id", default=None)
    value: ipaddress.IPv4Address | ipaddress.IPv6Address


class IPSet(BaseXmlModel):
    ip: list[IP] | None = element(tag="IP", default=None)
    ip_range: list[IPRange] | None = element(tag="IP_RANGE", default=None)


class AssetGroup(BaseXmlModel):
    id: int = element(tag="ID")
    title: str | None = element(tag="TITLE", default=None)
    owner_user_id: str | None = element(tag="OWNER_USER_ID", default=None)
    owner_unit_id: str | None = element(tag="OWNER_UNIT_ID", default=None)
    network_id: str | None = element(tag="NETWORK_ID", default=None)
    network_ids: str | None = element(tag="NETWORK_IDS", default=None)
    last_update: datetime.datetime | None = element(tag="LAST_UPDATE", default=None)
    business_impact: str | None = element(tag="BUSINESS_IMPACT", default=None)
    cvss_enviro_cdp: str | None = element(tag="CVSS_ENVIRO_CDP", default=None)
    cvss_enviro_td: str | None = element(tag="CVSS_ENVIRO_TD", default=None)
    cvss_enviro_cr: str | None = element(tag="CVSS_ENVIRO_CR", default=None)
    cvss_enviro_ir: str | None = element(tag="CVSS_ENVIRO_IR", default=None)
    cvss_enviro_ar: str | None = element(tag="CVSS_ENVIRO_AR", default=None)
    default_appliance_id: str | None = element(tag="DEFAULT_APPLIANCE_ID", default=None)
    appliance_ids: str | None = element(tag="APPLIANCE_IDS", default=None)
    ip_set: IPSet | None = element(tag="IP_SET", default=None)
    domain_list: list[Domain] | None = wrapped(
        "DOMAIN_LIST", element(tag="DOMAIN", default=None)
    )
    dns_list: list[DNS] | None = wrapped("DNS_LIST", element(tag="DNS", default=None))
    netbios_list: list[NetBIOS] | None = wrapped(
        "NETBIOS_LIST", element(tag="NETBIOS", default=None)
    )
    host_ids: str | None = element(tag="HOST_IDS", default=None)
    ec2_ids: str | None = element(tag="EC2_IDS", default=None)
    assigned_user_ids: str | None = element(tag="ASSIGNED_USER_IDS", default=None)
    assigned_unit_ids: str | None = element(tag="ASSIGNED_UNIT_IDS", default=None)
    comments: str | None = element(tag="COMMENTS", default=None)
    owner_user_name: str | None = element(tag="OWNER_USER_NAME", default=None)


class IdSet(BaseXmlModel):
    id: list[int] | None = element(tag="ID", default=None)
    id_range: list[str] | None = element(tag="ID_RANGE", default=None)


class Response(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    asset_group_list: list[AssetGroup] | None = wrapped(
        "ASSET_GROUP_LIST", element(tag="ASSET_GROUP", default=None)
    )
    id_set: list[IdSet] | None = element(tag="ID_SET", default=None)
    warning: ResponseWarning | None = element(tag="WARNING", default=None)


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    request_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] = wrapped("PARAM_LIST", element(tag="PARAM", default=None))
    post_data: str | None = element(tag="POST_DATA", default=None)


class AssetGroupListOutput(BaseXmlModel, tag="ASSET_GROUP_LIST_OUTPUT"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE")



import datetime
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic.alias_generators import to_camel
from pydantic.networks import IPvAnyAddress


class Model(BaseModel):
    model_config = ConfigDict(validate_by_name=True, alias_generator=to_camel)


class ProtocolSupportInfo(Model):
    sslv2: bool = Field(alias="SSLv2")
    sslv3: bool = Field(alias="SSLv3")
    tlsv1: bool = Field(alias="TLSv1")
    tlsv1_1: bool = Field(alias="TLSv1.1")
    tlsv1_2: bool = Field(alias="TLSv1.2")
    tlsv1_3: bool = Field(alias="TLSv1.3")


class CipherStrengthInfo(Model):
    less_than_128: bool = Field(alias="< 128 bits (e.g., 40, 56)")
    zero_bits: bool = Field(alias="0 bits (no encryption)")
    greater_equal_256: bool = Field(alias=">= 256 bits (e.g., 256)")
    less_than_256: bool = Field(alias="< 256 bits (e.g., 128, 168)")


class KeyExchangeInfo(BaseModel):
    key_or_dh_param_strength_lt_2048: bool = Field(
        alias="Key or DH parameter strength < 2048 bits (e.g., 1024)"
    )
    key_or_dh_param_strength_lt_512: bool = Field(
        alias="Key or DH parameter strength < 512 bits"
    )
    weak_key_debian_openssl: bool = Field(alias="Weak key (Debian OpenSSL flaw)")
    key_or_dh_param_strength_lt_4096: bool = Field(
        alias="Key or DH parameter strength < 4096 bits (e.g., 2048)"
    )
    key_or_dh_param_strength_ge_4096: bool = Field(
        alias="Key or DH parameter strength >= 4096 bits (e.g., 4096)"
    )
    exportable_key_exchange: bool = Field(alias="Exportable key exchange")
    anonymous_key_exchange: bool = Field(
        alias="Anonymous key exchange (no authentication)"
    )
    key_or_dh_param_strength_lt_1024: bool = Field(
        alias="Key or DH parameter strength < 1024 bits (e.g., 512)"
    )


class CipherSuite(Model):
    name: str
    key_exchange: str
    encryption_key_strength: int
    category: str


class CipherSuites(Model):
    sslv2: list[CipherSuite] | None = Field(alias="SSLv2", default=None)
    sslv3: list[CipherSuite] | None = Field(alias="SSLv3", default=None)
    tlsv1: list[CipherSuite] | None = Field(alias="TLSv1", default=None)
    tlsv1_1: list[CipherSuite] | None = Field(alias="TLSv1.1", default=None)
    tlsv1_2: list[CipherSuite] | None = Field(alias="TLSv1.2", default=None)
    tlsv1_3: list[CipherSuite] | None = Field(alias="TLSv1.3", default=None)


class Asset(Model):
    id: int
    uuid: str
    name: str
    primary_ip: IPvAnyAddress


class Certificate(Model):
    id: int
    certhash: str
    name: str
    last_found: datetime.datetime

    @field_validator("last_found", mode="after")
    @classmethod
    def parse_last_found(cls, value: Any) -> Any:
        if isinstance(value, (int, float)):
            return datetime.datetime.fromtimestamp(value)
        return value


class GradeSummary(Model):
    grade: str
    grade_with_trust_ignored: str
    certificate_score: int
    protocol_support_score: int
    key_exchange_score: int
    cipher_strength_score: int
    warnings: list[str]
    errors: list[str]
    notices: list[str]
    infos: list[str]
    highlights: list[str]
    protocol_support_info: ProtocolSupportInfo
    protocol_support_weightage: int
    cipher_strength_info: CipherStrengthInfo
    cipher_strength_weightage: int
    key_exchange_info: KeyExchangeInfo
    key_exchange_weightage: int
    cipher_suites: CipherSuites


class Instance(Model):
    id: int
    port: int
    scanned_date: datetime.datetime
    protocol: str
    service: str
    grade: str
    asset: Asset
    certificate: Certificate
    grade_summary: GradeSummary



# mypy: ignore-errors
# type: ignore

import inspect
import ipaddress
import os
import sys
import unittest

import sqlalchemy as sa

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from qualyspy import vmdr  # noqa: E402
from qualyspy.models.vmdr import host_list_orm, host_list_vm_detection_orm  # noqa: E402


class TestOutputModels(unittest.TestCase):
    def test_host_list(self):
        api = vmdr.VmdrAPI()
        host_list, _, _ = api.host_list(
            ids=11619472, show_trurisk=True, show_trurisk_factors=True
        )
        host = host_list[0]

        self.assertEqual(host.ip, ipaddress.ip_address("172.16.76.84"))

    def test_host_list_by_ip(self):
        api = vmdr.VmdrAPI()
        ips = ["172.16.68.91", "129.97.128.5", "129.97.85.51", "172.16.64.51"]
        host_list, _, _ = api.host_list(ips=ips, show_tags=True)
        host = host_list[0]

        self.assertEqual(host.ip, ipaddress.ip_address("172.16.68.91"))

    def test_host_list_vm_detection(self):
        api = vmdr.VmdrAPI()
        host_list, _, _ = api.host_list_vm_detection(ids=11619472, show_igs=True)
        host = host_list[0]

        self.assertEqual(host.ip, ipaddress.ip_address("172.16.76.84"))

    def test_knowledgebase(self):
        api = vmdr.VmdrAPI()
        kb = api.knowledgebase(ids=92203)
        vuln = kb[0]

        self.assertEqual(vuln.qid, 92203)

    def test_launch_vm_scan(self):
        api = vmdr.VmdrAPI()
        ret = api.launch_vm_scan(
            scan_title="Test Scan",
            iscanner_name="iss-qg-cr-v05",
            option_title="CertViewFree Profile",
            fqdn="learn.uwaterloo.ca",
        )
        text = ret.response.text

        self.assertEqual(text, "New vm scan launched")

    def test_vm_scan_list(self):
        api = vmdr.VmdrAPI()
        resp = api.vm_scan_list()
        scan = resp.response.scan_list[0]

        self.assertTrue(scan.ref.startswith("scan/"))

    def test_map_report_list(self):
        api = vmdr.VmdrAPI()
        reports = api.map_report_list(last=True)
        report = reports.report_list[0]

        self.assertEqual(report.title, "External Map baseline")

    def test_map_report(self):
        api = vmdr.VmdrAPI()
        reports = api.map_report_list(last=True)
        report_ref = reports.report_list[0].ref

        report = api.download_saved_map_report(ref=report_ref)

        self.assertEqual(report.value, report_ref)

    def test_asset_group_list(self):
        api = vmdr.VmdrAPI()
        asset_groups = api.asset_group_list()
        asset_group_names = [ag.title for ag in asset_groups]

        self.assertIn("Health", asset_group_names)

    def test_edit_asset_group(self):
        api = vmdr.VmdrAPI()
        asset_groups = api.asset_group_list()
        test_group = next(
            (ag for ag in asset_groups if ag.title == "QualysPy Test"), None
        )
        if test_group:
            edit = api.edit_asset_group(
                id=test_group.id,
                set_ips=[
                    "129.97.83.104",
                    ipaddress.ip_address("172.16.76.49"),
                    ipaddress.ip_network("172.19.15.0/27"),
                ],
            )

            self.assertEqual(edit.response.text, "Asset Group Updated Successfully")


class TestORM(unittest.TestCase):
    def test_orm_host_list(self):
        api = vmdr.HostListORM()
        api.load_safe(show_tags=True)
        stmt = sa.select(host_list_orm.Host).where(host_list_orm.Host.id == 11619472)
        result = api.query(stmt)
        host = result[0][0]
        self.assertEqual(host.ip, ipaddress.ip_address("172.16.76.84"))

    def test_orm_vm_detection(self):
        api = vmdr.HostListVMDetectionORM()
        api.drop()
        api.init_db()
        api.load(show_igs=True)
        stmt = sa.select(host_list_vm_detection_orm.Host).where(
            host_list_vm_detection_orm.Host.id == 11619472
        )
        result = api.query(stmt)
        host = result[0][0]
        self.assertEqual(host.ip, ipaddress.ip_address("172.16.76.84"))

    def test_orm_knowledgebase(self):
        api = vmdr.KnowledgebaseORM()
        api.drop()
        api.init_db()
        api.load(details="All")
        stmt = sa.select(vmdr.knowledgebase_orm.Vuln).where(
            vmdr.knowledgebase_orm.Vuln.qid == 6
        )
        result = api.query(stmt)
        vuln = result[0][0]
        self.assertEqual(vuln.title, "DNS Host Name")


if __name__ == "__main__":
    unittest.main()



import datetime

import sqlalchemy as sa
import sqlalchemy.orm as orm

from .. import sa_types


class Base(orm.DeclarativeBase):
    metadata = sa.MetaData(schema="asset_details")


class CipherSuite(Base):
    __tablename__ = "cipher_suite"

    name: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    key_exchange: orm.Mapped[str]
    encryption_key_strength: orm.Mapped[int]
    category: orm.Mapped[str]


class CipherSuites(Base):
    __tablename__ = "cipher_suites"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    sslv3: orm.Mapped[list["CipherSuite"]] = orm.relationship(
        back_populates="sslv3", uselist=True
    )
    tlsv1: orm.Mapped[list["CipherSuite"]] = orm.relationship(
        back_populates="tlsv1", uselist=True
    )
    tlsv1_1: orm.Mapped[list["CipherSuite"]] = orm.relationship(
        back_populates="tlsv1_1", uselist=True
    )
    tlsv1_2: orm.Mapped[list["CipherSuite"]] = orm.relationship(
        back_populates="tlsv1_2", uselist=True
    )
    tlsv1_3: orm.Mapped[list["CipherSuite"]] = orm.relationship(
        back_populates="tlsv1_3", uselist=True
    )


class Asset(Base):
    __tablename__ = "asset"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    uuid: orm.Mapped[str]
    name: orm.Mapped[str]
    primary_ip: orm.Mapped[sa_types.IPAddressGenericType]

    instances: orm.Mapped[list["Instance"]] = orm.relationship(
        back_populates="asset", uselist=True
    )


class Certificate(Base):
    __tablename__ = "certificate"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    certhash: orm.Mapped[str]
    name: orm.Mapped[str]
    last_found: orm.Mapped[datetime.datetime]

    instances: orm.Mapped[list["Instance"]] = orm.relationship(
        back_populates="certificate", uselist=True
    )


class GradeSummary(Base):
    __tablename__ = "grade_summary"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    grade: orm.Mapped[str]
    grade_with_trust_ignored: orm.Mapped[str]
    certificate_score: orm.Mapped[int]
    protocol_support_score: orm.Mapped[int]
    key_exchange_score: orm.Mapped[int]
    cipher_strength_score: orm.Mapped[int]
    warnings: orm.Mapped[list[str]]
    errors: orm.Mapped[list[str]]
    notices: orm.Mapped[list[str]]
    infos: orm.Mapped[list[str]]
    highlights: orm.Mapped[list[str]]
    protocol_support_info: orm.Mapped[ProtocolSupportInfo]
    protocol_support_weightage: orm.Mapped[int]
    cipher_strength_info: orm.Mapped[CipherStrengthInfo]
    cipher_strength_weightage: orm.Mapped[int]
    key_exchange_info: orm.Mapped[KeyExchangeInfo]
    key_exchange_weightage: orm.Mapped[int]
    cipher_suites: orm.Mapped[CipherSuites]


class Instance(Base):
    __tablename__ = "instance"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    port: orm.Mapped[int]
    scanned_date: orm.Mapped[datetime.datetime]
    protocol: orm.Mapped[str]
    service: orm.Mapped[str]
    grade: orm.Mapped[str]
    asset_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("asset.id"))
    asset: orm.Mapped[Asset] = orm.relationship(
        back_populates="instances", uselist=False
    )
    certificate_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("certificate.id"))
    certificate: orm.Mapped[Certificate] = orm.relationship(
        back_populates="instances", uselist=False
    )
    grade_summary: orm.Mapped[GradeSummary] = orm.relationship(
        back_populates="instance", uselist=False
    )



"""Exceptions for QualysPy."""


class QualysAPIError(Exception):
    """Indicates an error with the Qualys API."""

    pass


class ConfigError(Exception):
    """Indicates an error with the configuration file."""

    pass


class ValidationError(Exception):
    """Wrapper for pydantic.error_wrappers.ValidationError"""

    pass


class TimeoutError(Exception):
    """Wrapper for requests.exceptions.Timeout"""

    pass



import logging
import pathlib
from logging.handlers import RotatingFileHandler

from decouple import config  # type: ignore

def bootstrap_logger() -> logging.Logger:
    enabled = config("QUALYSPY_LOG_ENABLED", default="False", cast=bool)
    log_file = str(config("QUALYSPY_LOG_FILE", default=pathlib.Path.home() / "qualyspy.log"))
    level_name = str(config("QUALYSPY_LOG_LEVEL", default="INFO")).upper()
    rotate_mb = config("QUALYSPY_LOG_MAX_MB", default=10, cast=int)
    backups = config("QUALYSPY_LOG_BACKUPS", default=5, cast=int)
    datefmt = "%Y-%m-%d %H:%M:%S"


    logger = logging.getLogger("qualyspy")
    if logger.handlers:  # protect against double-initialisation
        return logger
    
    if not enabled:
        logger.addHandler(logging.NullHandler())
        logger.setLevel(logging.CRITICAL)    # silence by default
        logger.propagate = False
        return logger

    logger.setLevel(level_name)

    if rotate_mb == 0:  # allow “no rotation” by env
        handler = logging.FileHandler(log_file, encoding="utf-8")
    else:
        handler = RotatingFileHandler(
            log_file,
            maxBytes=rotate_mb * 1024 * 1024,
            backupCount=backups,
            encoding="utf-8",
        )

    fmt = (
        "%(asctime)s | %(levelname)s | %(name)s | %(funcName)s:%(lineno)d | %(message)s"
    )
    handler.setFormatter(logging.Formatter(fmt, datefmt))
    logger.addHandler(handler)
    logger.propagate = False  # don’t spam root logger
    return logger



"""Utility functions for qualyspy.  Primarily for internal use."""

import copy
import dataclasses
import importlib
import inspect
import re
from typing import Any, Sequence, TypeVar

from sqlalchemy import inspect as sqlalchemy_inspect

_D = TypeVar("_D")
_RE_QUALYSPY_CLASSNAME = re.compile(r"(qualyspy[\w._]*)")
_RE_SA_CLASSNAME = re.compile(r"sqlalchemy.orm")


def _get_cls_inst_from_annot(mapped_cls: str) -> Any:
    """Get an instance of a class from a string annotation.

    Args:
        mapped_cls (str): The string annotation of the class.

    Returns:
        Any: An instance of the class.

    Raises:
        ValueError: If the annotation is not a Mapped class.
    """
    m = _RE_QUALYSPY_CLASSNAME.search(mapped_cls)
    if m is not None:
        child_cls_strs = m.group(1).split(".")
        child_cls_package = ".".join(child_cls_strs[:-1])
        mod = importlib.import_module(child_cls_package)
        child_cls = getattr(mod, child_cls_strs[-1])
        return child_cls
    else:
        if _RE_SA_CLASSNAME.search(mapped_cls):
            return None
        else:
            raise ValueError("Annotation is not a Mapped class.")


def to_orm_object(
    obj: Any,
    out_cls: type[_D],
) -> _D:
    """Convert a dataclass instance of a Qualys object to an ORM object.

    Args:
        obj (dict[str, Any]): The dataclass instance of a Qualys object.
        out_cls (type[_D]): The ORM class to convert to.

    Returns:
        _D: The ORM object.
    """

    def _to_orm_object(
        obj: dict[str, Any],
        out_cls: type[_D],
    ) -> _D:
        """Helper function for to_orm_object.  Recursively converts a dataclass instance of a Qualys
        object to an ORM object.  This separate function allows obj to be converted from a
        dataclass to a dict before the recursive calls."""

        obj_copy = copy.deepcopy(obj)
        annots = inspect.get_annotations(out_cls)
        for k, v in obj_copy.items():
            if isinstance(v, dict):
                mapped_cls = str(annots[k])
                child_cls = _get_cls_inst_from_annot(mapped_cls)
                if child_cls is None:
                    continue
                obj_copy[k] = _to_orm_object(v, child_cls)
            elif isinstance(v, list) and len(v) > 0:
                mapped_cls = str(annots[k])
                child_cls = _get_cls_inst_from_annot(mapped_cls)
                if child_cls is None:
                    continue
                if isinstance(v[0], dict):
                    v = [_to_orm_object(item, child_cls) for item in v]
                else:
                    child_annots = inspect.get_annotations(child_cls)
                    param = next(iter(child_annots))
                    v = [child_cls(**{param: item}) for item in v]
                obj_copy[k] = v

        return out_cls(**obj_copy)

    try:  # Type is Pydantic dataclass
        obj_dict = dataclasses.asdict(obj)
    except TypeError:  # Type is Pydantic model
        obj_dict = obj.dict()
    return _to_orm_object(obj_dict, out_cls)


def to_orm_objects(objs: Sequence[Any], out_cls: type[_D]) -> list[_D]:
    obj_cache: dict[type[Any], dict[tuple[Any, ...], Any]] = {}

    def get_primary_key_values(
        cls: type[Any], inst: dict[str, Any]
    ) -> tuple[Any, ...]:
        mapper = sqlalchemy_inspect(cls)
        if mapper is None:
            raise ValueError("Class is not a mapped class.")

        pk_names = [col.name for col in mapper.primary_key]

        return tuple(inst.get(pk_name) for pk_name in pk_names)

    def to_orm_object(obj: dict[str, Any], out_cls: type[_D]) -> _D:
        nonlocal obj_cache

        out_cls_cache = obj_cache.get(out_cls)
        if out_cls_cache is not None:
            out_cls_cache_key = get_primary_key_values(out_cls, obj)
            if out_cls_cache_key is not None:
                orm_obj = out_cls_cache.get(out_cls_cache_key)
                if orm_obj is not None:
                    return orm_obj  # type: ignore

        obj_copy = copy.deepcopy(obj)
        annots = inspect.get_annotations(out_cls)
        for k, v in obj_copy.items():
            if isinstance(v, dict):
                mapped_cls = str(annots[k])
                child_cls = _get_cls_inst_from_annot(mapped_cls)
                if child_cls is None:
                    continue
                obj_copy[k] = to_orm_object(v, child_cls)
            elif isinstance(v, list) and len(v) > 0:
                mapped_cls = str(annots[k])
                child_cls = _get_cls_inst_from_annot(mapped_cls)
                if child_cls is None:
                    continue
                if isinstance(v[0], dict):
                    v = [to_orm_object(item, child_cls) for item in v]
                else:
                    child_annots = inspect.get_annotations(child_cls)
                    param = next(iter(child_annots))
                    v = [child_cls(**{param: item}) for item in v]
                obj_copy[k] = v

        orm_obj = out_cls(**obj_copy)
        primary_key_values = get_primary_key_values(out_cls, obj)
        if primary_key_values != (None,):
            if out_cls_cache is None:
                obj_cache[out_cls] = {primary_key_values: orm_obj}
            else:
                out_cls_cache[primary_key_values] = orm_obj
        return orm_obj

    orm_objs = []
    for obj in objs:
        try:
            obj_dict = dataclasses.asdict(obj)
        except TypeError:  # Type is Pydantic model
            obj_dict = obj.dict()
        orm_obj = to_orm_object(obj_dict, out_cls)
        orm_objs.append(orm_obj)

    return orm_objs


def from_orm_object(obj: Any, output_class: Any) -> Any:
    """Convert an ORM object to a dataclass instance of a Qualys object.

    Args:
        obj (Any): The ORM object.

    Returns:
        Any: The dataclass instance of a Qualys object.
    """

    # Since the ORM objects relationships are bidirectional, we need to keep track of which
    # classes we've already processed to avoid infinite recursion.
    processed_classes: set[Any] = set()

    builtin_types = (str, int, float, bool, type(None))

    def get_attributes(obj: Any) -> dict[str, Any]:
        """Get the attributes of an ORM object.

        Args:
            obj (Any): The ORM object.

        Returns:
            dict[str, Any]: The attributes of the ORM object.
        """
        nonlocal processed_classes
        processed_classes.add(type(obj))

        result: dict[str, Any] = {}
        keys = obj.__mapper__.attrs.keys()  # List of attributes of the ORM object
        for key in keys:
            value = getattr(obj, key)

            # Check for bultin types first so they don't get added to processed_classes
            if isinstance(value, builtin_types):
                result[key] = value
                continue

            # Skip classes we've already processed so we don't go back up the hierarchy
            if type(value) in processed_classes:
                continue
            else:
                processed_classes.add(type(value))

            if callable(value):  # Skip methods
                continue
            elif isinstance(value, list):
                result[key] = [_from_orm_object(v) for v in value]
            else:
                result[key] = _from_orm_object(value)
        return result

    def _from_orm_object(
        obj: Any,
    ) -> Any:
        """Helper function for from_orm_object.  Recursively converts an ORM object to a dataclass
        instance of a Qualys object.
        """
        if isinstance(obj, list):
            return [_from_orm_object(item) for item in obj]
        elif not hasattr(obj, "__dict__"):
            return obj
        return get_attributes(obj)

    return output_class(**_from_orm_object(obj))


def snake_to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # capitalize the first component and join the rest
    return components[0] + "".join(x.title() for x in components[1:])


def clean_dict(d: dict[str, Any]) -> dict[str, str]:
    """Remove None values from a dictionary and convert values to strings.

    Args:
        d (dict[str, Any]): The dictionary to clean.

    Returns:
        dict[str, str]: The cleaned dictionary.
    """

    def _clean_dict(v: Any) -> str:
        # if l is a list, return a comma-separated string of the list items
        if isinstance(v, list):
            return ",".join([str(item) for item in v])
        elif isinstance(v, bool):
            return "1" if v else "0"
        else:
            return str(v)

    return {k: _clean_dict(v) for k, v in d.items() if v is not None}



from pydantic_xml import BaseXmlModel, attr, element, wrapped


class OptionProfile(BaseXmlModel):
    option_profile_title: str = element(tag="OPTION_PROFILE_TITLE")
    option_profile_default: str | None = attr(
        name="option_profile_default", default=None
    )


class AssetGroup(BaseXmlModel):
    asset_group_title: str = element(tag="ASSET_GROUP_TITLE")


class MapReport(BaseXmlModel):
    title: str = element(tag="TITLE")
    asset_groups: list[AssetGroup] | None = wrapped(
        "ASSET_GROUPS", element(tag="ASSET_GROUP", default=None)
    )
    option_profile: OptionProfile | None = element(tag="OPTION_PROFILE", default=None)
    ref: str = attr(name="ref")
    date: str = attr(name="date")
    domain: str = attr(name="domain")
    status: str = attr(name="status")


class Error(BaseXmlModel):
    number: int | None = element(tag="NUMBER", default=None)
    value: str


class MapReportList(BaseXmlModel, tag="MAP_REPORT_LIST"):
    report_list: list[MapReport] | None = element(tag="MAP_REPORT", default=None)
    error: Error | None = element(tag="ERROR", default=None)
    user: str = attr(name="user")
    from_: str = attr(name="from")
    to: str = attr(name="to")
    with_domain: str | None = attr(name="with_domain", default=None)



import datetime

import sqlalchemy as sa
import sqlalchemy.orm as orm


class Base(orm.DeclarativeBase):
    metadata = sa.MetaData(schema="knowledgebase")


class ChangeLog(Base):
    __tablename__ = "change_log"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    changes_date: orm.Mapped[datetime.datetime]
    comments: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="change_log_list")


class Discovery(Base):
    __tablename__ = "discovery"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    remote: orm.Mapped[bool]
    auth_type_list: orm.Mapped[list[str]] = orm.mapped_column(sa.ARRAY(sa.String))
    additional_info: orm.Mapped[str | None]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="discovery")


class ThreatIntel(Base):
    __tablename__ = "threat_intel"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    threat_intel_id: orm.Mapped[str]
    value: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="threat_intelligence")


class Attack(Base):
    __tablename__ = "attack"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    vector: orm.Mapped[str | None]
    complexity: orm.Mapped[str | None]

    cvss_v3_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss_v3.id"))
    cvss_v3: orm.Mapped["CVSSv3"] = orm.relationship(back_populates="attack")


class CVSSv3Impact(Base):
    __tablename__ = "cvss_v3_impact"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    confidentiality: orm.Mapped[str | None]
    integrity: orm.Mapped[str | None]
    availability: orm.Mapped[str | None]

    cvss_v3_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss_v3.id"))
    cvss_v3: orm.Mapped["CVSSv3"] = orm.relationship(back_populates="impact")


class CVSSImpact(Base):
    __tablename__ = "cvss_impact"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    confidentiality: orm.Mapped[str | None]
    integrity: orm.Mapped[str | None]
    availability: orm.Mapped[str | None]

    cvss_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss.id"))
    cvss: orm.Mapped["CVSS"] = orm.relationship(back_populates="impact")


class Access(Base):
    __tablename__ = "access"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    vector: orm.Mapped[str | None]
    complexity: orm.Mapped[str | None]

    cvss_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss.id"))
    cvss: orm.Mapped["CVSS"] = orm.relationship(back_populates="access")


class CVSSv3Base(Base):
    __tablename__ = "cvss_v3_base"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    source: orm.Mapped[str | None]
    value: orm.Mapped[str]

    cvss_v3_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss_v3.id"))
    cvss_v3: orm.Mapped["CVSSv3"] = orm.relationship(back_populates="base")


class CVSSBase(Base):
    __tablename__ = "cvss_base"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    source: orm.Mapped[str | None]
    value: orm.Mapped[str]

    cvss_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("cvss.id"))
    cvss: orm.Mapped["CVSS"] = orm.relationship(back_populates="base")


class CVSSv3(Base):
    __tablename__ = "cvss_v3"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    base: orm.Mapped[CVSSv3Base | None] = orm.relationship(
        back_populates="cvss_v3", uselist=False
    )
    temporal: orm.Mapped[float | None]
    vector_string: orm.Mapped[str | None]
    cvss3_version: orm.Mapped[str | None]
    attack: orm.Mapped[Attack | None] = orm.relationship(
        back_populates="cvss_v3", uselist=False
    )
    impact: orm.Mapped[CVSSv3Impact | None] = orm.relationship(
        back_populates="cvss_v3", uselist=False
    )
    privileges_required: orm.Mapped[str | None]
    user_interaction: orm.Mapped[str | None]
    scope: orm.Mapped[str | None]
    exploit_code_maturity: orm.Mapped[str | None]
    remediation_level: orm.Mapped[str | None]
    report_confidence: orm.Mapped[str | None]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="cvss_v3")


class CVSS(Base):
    __tablename__ = "cvss"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    base: orm.Mapped[CVSSBase | None] = orm.relationship(
        back_populates="cvss", uselist=False
    )
    temporal: orm.Mapped[float | None]
    vector_string: orm.Mapped[str | None]
    cvss3_version: orm.Mapped[str | None]
    access: orm.Mapped[Access | None] = orm.relationship(
        back_populates="cvss", uselist=False
    )
    impact: orm.Mapped[CVSSImpact | None] = orm.relationship(
        back_populates="cvss", uselist=False
    )
    authentication: orm.Mapped[str | None]
    exploitability: orm.Mapped[str | None]
    remediation_level: orm.Mapped[str | None]
    report_confidence: orm.Mapped[str | None]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="cvss")


class MWInfo(Base):
    __tablename__ = "mw_info"

    mw_id: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    mw_type: orm.Mapped[str | None]
    mw_platform: orm.Mapped[str | None]
    mw_alias: orm.Mapped[str | None]
    mw_rating: orm.Mapped[str | None]
    mw_link: orm.Mapped[str | None]

    mw_src_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("mw_src.id"))
    mw_src: orm.Mapped["MWSrc"] = orm.relationship(back_populates="mw_list")


class MWSrc(Base):
    __tablename__ = "mw_src"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    src_name: orm.Mapped[str]
    mw_list: orm.Mapped[list[MWInfo]] = orm.relationship(
        back_populates="mw_src", uselist=True
    )

    correlation_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("correlation.id"))
    correlation: orm.Mapped["Correlation"] = orm.relationship(back_populates="malware")


class Explt(Base):
    __tablename__ = "explt"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ref: orm.Mapped[str]
    desc: orm.Mapped[str]
    link: orm.Mapped[str | None]

    explt_src_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("explt_src.id"))
    explt_src: orm.Mapped["ExpltSrc"] = orm.relationship(back_populates="explt_list")


class ExpltSrc(Base):
    __tablename__ = "explt_src"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    src_name: orm.Mapped[str | None]
    explt_list: orm.Mapped[list[Explt]] = orm.relationship(
        back_populates="explt_src", uselist=True
    )

    correlation_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("correlation.id"))
    correlation: orm.Mapped["Correlation"] = orm.relationship(back_populates="exploits")


class Correlation(Base):
    __tablename__ = "correlation"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    exploits: orm.Mapped[list[ExpltSrc]] = orm.relationship(
        back_populates="correlation", uselist=True
    )
    malware: orm.Mapped[list[MWSrc]] = orm.relationship(
        back_populates="correlation", uselist=True
    )

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="correlation")


class Compliance(Base):
    __tablename__ = "compliance"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    type: orm.Mapped[str]
    section: orm.Mapped[str]
    description: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="compliance_list")


cve_association_table = sa.Table(
    "cve_association",
    Base.metadata,
    sa.Column("cve_id", sa.ForeignKey("cve.id")),
    sa.Column("vuln_qid", sa.ForeignKey("vuln.qid")),
)


class CVE(Base):
    __tablename__ = "cve"

    id: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    url: orm.Mapped[str]

    vulns: orm.Mapped[list["Vuln"]] = orm.relationship(
        secondary=cve_association_table, back_populates="cve_list", uselist=True
    )


class VendorReference(Base):
    __tablename__ = "vendor_reference"

    id: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    url: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="vendor_reference_list")


class Software(Base):
    __tablename__ = "software"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    product: orm.Mapped[str]
    vendor: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="software_list")


class Bugtraq(Base):
    __tablename__ = "bugtraq"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    bugtraq_id: orm.Mapped[str]
    url: orm.Mapped[str]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="bugtraq_list")


class LastCustomization(Base):
    __tablename__ = "last_customization"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    last_customization_datetime: orm.Mapped[datetime.datetime | None]
    user_login: orm.Mapped[str | None]

    vuln_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("vuln.qid"))
    vuln: orm.Mapped["Vuln"] = orm.relationship(back_populates="last_customization")


class Vuln(Base):
    __tablename__ = "vuln"

    qid: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    vuln_type: orm.Mapped[str]
    severity_level: orm.Mapped[int]
    title: orm.Mapped[str]
    category: orm.Mapped[str | None]
    detection_info: orm.Mapped[str | None]
    last_customization: orm.Mapped[LastCustomization | None] = orm.relationship(
        back_populates="vuln", uselist=False
    )
    last_service_modification_datetime: orm.Mapped[datetime.datetime | None]
    published_datetime: orm.Mapped[datetime.datetime | None]
    code_modified_datetime: orm.Mapped[datetime.datetime | None]
    bugtraq_list: orm.Mapped[list[Bugtraq]] = orm.relationship(
        back_populates="vuln", uselist=True
    )
    patchable: orm.Mapped[bool | None]
    patch_published_date: orm.Mapped[datetime.datetime | None]
    software_list: orm.Mapped[list[Software]] = orm.relationship(
        back_populates="vuln", uselist=True
    )
    vendor_reference_list: orm.Mapped[list[VendorReference]] = orm.relationship(
        back_populates="vuln", uselist=True
    )
    cve_list: orm.Mapped[list[CVE]] = orm.relationship(
        secondary=cve_association_table, back_populates="vulns", uselist=True
    )
    diagnosis: orm.Mapped[str | None]
    diagnosis_comment: orm.Mapped[str | None]
    consequence: orm.Mapped[str | None]
    consequence_comment: orm.Mapped[str | None]
    solution: orm.Mapped[str | None]
    solution_comment: orm.Mapped[str | None]
    compliance_list: orm.Mapped[list[Compliance]] = orm.relationship(
        back_populates="vuln", uselist=True
    )
    correlation: orm.Mapped[Correlation | None] = orm.relationship(
        back_populates="vuln", uselist=False
    )
    cvss: orm.Mapped[CVSS | None] = orm.relationship(
        back_populates="vuln", uselist=False
    )
    cvss_v3: orm.Mapped[CVSSv3 | None] = orm.relationship(
        back_populates="vuln", uselist=False
    )
    pci_flag: orm.Mapped[bool | None]
    automatic_pci_fail: orm.Mapped[bool | None]
    pci_reasons: orm.Mapped[list[str]] = orm.mapped_column(sa.ARRAY(sa.String))
    threat_intelligence: orm.Mapped[list[ThreatIntel]] = orm.relationship(
        back_populates="vuln", uselist=True
    )
    supported_modules: orm.Mapped[str | None]
    discovery: orm.Mapped[Discovery | None] = orm.relationship(
        back_populates="vuln", uselist=False
    )
    is_disabled: orm.Mapped[bool | None]
    change_log_list: orm.Mapped[list[ChangeLog]] = orm.relationship(
        back_populates="vuln", uselist=True
    )



import datetime

from pydantic_xml import BaseXmlModel, attr, element, wrapped


class ResponseWarning(BaseXmlModel):
    code: str = element(tag="CODE")
    text: str = element(tag="TEXT")
    url: str = element(tag="URL")


class ChangeLog(BaseXmlModel):
    changes_date: datetime.datetime = element(tag="CHANGES_DATE")
    comments: str = element(tag="COMMENTS")


class Discovery(BaseXmlModel):
    remote: bool = element(tag="REMOTE")
    auth_type_list: list[str] = wrapped(
        "AUTH_TYPE_LIST", element(tag="AUTH_TYPE", default_factory=list)
    )
    additional_info: str | None = element(tag="ADDITIONAL_INFO", default=None)


class ThreatIntel(BaseXmlModel):
    threat_intel_id: str = attr(name="id")
    value: str


class Attack(BaseXmlModel):
    vector: str | None = element(tag="VECTOR", default=None)
    complexity: str | None = element(tag="COMPLEXITY", default=None)


class Impact(BaseXmlModel):
    confidentiality: str | None = element(tag="CONFIDENTIALITY", default=None)
    integrity: str | None = element(tag="INTEGRITY", default=None)
    availability: str | None = element(tag="AVAILABILITY", default=None)


class Access(BaseXmlModel):
    vector: str | None = element(tag="VECTOR", default=None)
    complexity: str | None = element(tag="COMPLEXITY", default=None)


class Base(BaseXmlModel):
    source: str | None = attr(name="source", default=None)
    value: str


class CVSSv3(BaseXmlModel):
    base: Base | None = element(tag="BASE", default=None)
    temporal: float | None = element(tag="TEMPORAL", default=None)
    vector_string: str | None = element(tag="VECTOR_STRING", default=None)
    cvss3_version: str | None = element(tag="CVSS3_VERSION", default=None)
    attack: Attack | None = element(tag="ATTACK", default=None)
    impact: Impact | None = element(tag="IMPACT", default=None)
    privileges_required: str | None = element(tag="PRIVILEGES_REQUIRED", default=None)
    user_interaction: str | None = element(tag="USER_INTERACTION", default=None)
    scope: str | None = element(tag="SCOPE", default=None)
    exploit_code_maturity: str | None = element(
        tag="EXPLOIT_CODE_MATURITY", default=None
    )
    remediation_level: str | None = element(tag="REMEDIATION_LEVEL", default=None)
    report_confidence: str | None = element(tag="REPORT_CONFIDENCE", default=None)


class CVSS(BaseXmlModel):
    base: Base | None = element(tag="BASE", default=None)
    temporal: float | None = element(tag="TEMPORAL", default=None)
    vector_string: str | None = element(tag="VECTOR_STRING", default=None)
    cvss3_version: str | None = element(tag="CVSS3_VERSION", default=None)
    access: Access | None = element(tag="ACCESS", default=None)
    impact: Impact | None = element(tag="IMPACT", default=None)
    authentication: str | None = element(tag="AUTHENTICATION", default=None)
    exploitability: str | None = element(tag="EXPLOITABILITY", default=None)
    remediation_level: str | None = element(tag="REMEDIATION_LEVEL", default=None)
    report_confidence: str | None = element(tag="REPORT_CONFIDENCE", default=None)


class MWInfo(BaseXmlModel):
    mw_id: str = element(tag="MW_ID")
    mw_type: str | None = element(tag="MW_TYPE", default=None)
    mw_platform: str | None = element(tag="MW_PLATFORM", default=None)
    mw_alias: str | None = element(tag="MW_ALIAS", default=None)
    mw_rating: str | None = element(tag="MW_RATING", default=None)
    mw_link: str | None = element(tag="MW_LINK", default=None)


class MWSrc(BaseXmlModel):
    src_name: str = element(tag="SRC_NAME")
    mw_list: list[MWInfo] = wrapped(
        "MW_LIST", element(tag="MW_INFO", default_factory=list)
    )


class Explt(BaseXmlModel):
    ref: str = element(tag="REF")
    desc: str = element(tag="DESC")
    link: str | None = element(tag="LINK", default=None)


class ExpltSrc(BaseXmlModel):
    src_name: str | None = element(tag="SRC_NAME", default=None)
    explt_list: list[Explt] = wrapped(
        "EXPLT_LIST", element(tag="EXPLT", default_factory=list)
    )


class Correlation(BaseXmlModel):
    exploits: list[ExpltSrc] = wrapped(
        "EXPLOITS", element(tag="EXPLT_SRC", default_factory=list)
    )
    malware: list[MWSrc] = wrapped(
        "MALWARE", element(tag="MW_SRC", default_factory=list)
    )


class Compliance(BaseXmlModel):
    type: str = element(tag="TYPE")
    section: str = element(tag="SECTION")
    description: str = element(tag="DESCRIPTION")


class CVE(BaseXmlModel):
    id: str = element(tag="ID")
    url: str = element(tag="URL")


class VendorReference(BaseXmlModel):
    id: str = element(tag="ID")
    url: str = element(tag="URL")


class Software(BaseXmlModel):
    product: str = element(tag="PRODUCT")
    vendor: str = element(tag="VENDOR")


class Bugtraq(BaseXmlModel):
    bugtraq_id: str = element(tag="ID")
    url: str = element(tag="URL")


class LastCustomization(BaseXmlModel):
    last_customization_datetime: datetime.datetime | None = element(
        tag="DATETIME", defaul=None
    )
    user_login: str = element(tag="USER_LOGIN")


class Vuln(BaseXmlModel):
    qid: int = element(tag="QID")
    vuln_type: str = element(tag="VULN_TYPE")
    severity_level: int = element(tag="SEVERITY_LEVEL")
    title: str = element(tag="TITLE")
    category: str | None = element(tag="CATEGORY", default=None)
    detection_info: str | None = element(tag="DETECTION_INFO", default=None)
    last_customization: LastCustomization | None = element(
        tag="LAST_CUSTOMIZATION", default=None
    )
    last_service_modification_datetime: datetime.datetime | None = element(
        tag="LAST_SERVICE_MODIFICATION_DATETIME", default=None
    )
    published_datetime: datetime.datetime | None = element(
        tag="PUBLISHED_DATETIME", default=None
    )
    code_modified_datetime: datetime.datetime | None = element(
        tag="CODE_MODIFIED_DATETIME", default=None
    )
    bugtraq_list: list[Bugtraq] | None = wrapped(
        "BUGTRAQ_LIST", element(tag="BUGTRAQ", default_factory=list)
    )
    patchable: bool | None = element(tag="PATCHABLE", default=None)
    patch_published_date: datetime.datetime | None = element(
        tag="PATCH_PUBLISHED_DATE", default=None
    )
    software_list: list[Software] = wrapped(
        "SOFTWARE_LIST", element(tag="SOFTWARE", default_factory=list)
    )
    vendor_reference_list: list[VendorReference] = wrapped(
        "VENDOR_REFERENCE_LIST", element(tag="VENDOR_REFERENCE", default_factory=list)
    )
    cve_list: list[CVE] = wrapped("CVE_LIST", element(tag="CVE", default_factory=list))
    diagnosis: str | None = element(tag="DIAGNOSIS", default=None)
    diagnosis_comment: str | None = element(tag="DIAGNOSIS_COMMENT", default=None)
    consequence: str | None = element(tag="CONSEQUENCE", default=None)
    consequence_comment: str | None = element(tag="CONSEQUENCE_COMMENT", default=None)
    solution: str | None = element(tag="SOLUTION", default=None)
    solution_comment: str | None = element(tag="SOLUTION_COMMENT", default=None)
    compliance_list: list[Compliance] = wrapped(
        "COMPLIANCE_LIST", element(tag="COMPLIANCE", default_factory=list)
    )
    correlation: Correlation | None = element(tag="CORRELATION", default=None)
    cvss: CVSS | None = element(tag="CVSS", default=None)
    cvss_v3: CVSSv3 | None = element(tag="CVSS_V3", default=None)
    pci_flag: bool | None = element(tag="PCI_FLAG", default=None)
    automatic_pci_fail: bool | None = element(tag="AUTOMATIC_PCI_FAIL", default=None)
    pci_reasons: list[str] = wrapped(
        "PCI_REASONS", element(tag="PCI_REASON", default_factory=list)
    )
    threat_intelligence: list[ThreatIntel] = wrapped(
        "THREAT_INTELLIGENCE", element(tag="THREAT_INTEL", default_factory=list)
    )
    supported_modules: str | None = element(tag="SUPPORTED_MODULES", default=None)
    discovery: Discovery | None = element(tag="DISCOVERY", default=None)
    is_disabled: bool | None = element(tag="IS_DISABLED", default=None)
    change_log_list: list[ChangeLog] = wrapped(
        "CHANGE_LOG_LIST", element(tag="CHANGE_LOG", default_factory=list)
    )


class Response(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    vuln_list: list[Vuln] = wrapped(
        "VULN_LIST", element(tag="VULN", default_factory=list)
    )
    id_set: list[int] = wrapped("ID_SET", element(tag="ID", default_factory=list))
    warning: ResponseWarning | None = element(tag="WARNING", default=None)


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    request_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] = element(tag="PARAM_LIST", default_factory=list)
    post_data: str | None = element(tag="POST_DATA")


class KnowledgeBaseOutput(BaseXmlModel, tag="KNOWLEDGE_BASE_VULN_LIST_OUTPUT"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE")



import datetime

from pydantic_xml import BaseXmlModel, element, wrapped


class Item(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Response(BaseXmlModel):
    reponse_datetime: datetime.datetime = element(tag="DATETIME")
    code: str | None = element(tag="CODE", default=None)
    text: str = element(tag="TEXT")
    item_list: list[Item] | None = wrapped(
        "ITEM_LIST", element(tag="ITEM", default=None)
    )


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    request_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] | None = element(tag="PARAM_LIST")
    post_data: str | None = element(tag="POST_DATA")


class SimpleReturn(BaseXmlModel, tag="SIMPLE_RETURN"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE")



import datetime as dt
import ipaddress

import sqlalchemy as sa
import sqlalchemy.dialects.postgresql as sa_pg
import sqlalchemy.orm as orm

from .. import sa_types


class Base(orm.DeclarativeBase):
    metadata = sa.MetaData(schema="asset_details")


class OperatingSystemTaxonomy(Base):
    __tablename__ = "operating_system_taxonomy"

    pk: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    id: orm.Mapped[int | None]
    name: orm.Mapped[str]
    category1: orm.Mapped[str | None]
    category2: orm.Mapped[str | None]

    operating_system_id: orm.Mapped[int | None] = orm.mapped_column(
        sa.ForeignKey("operating_system.id")
    )
    operating_system: orm.Mapped["OperatingSystem | None"] = orm.relationship(
        back_populates="taxonomy"
    )


class HardwareTaxonomy(Base):
    __tablename__ = "hardware_taxonomy"

    pk: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    id: orm.Mapped[int | None]
    name: orm.Mapped[str]
    category1: orm.Mapped[str | None]
    category2: orm.Mapped[str | None]

    hardware_id: orm.Mapped[int | None] = orm.mapped_column(
        sa.ForeignKey("hardware.id")
    )
    hardware: orm.Mapped["Hardware | None"] = orm.relationship(
        back_populates="taxonomy"
    )


class OperatingSystem(Base):
    __tablename__ = "operating_system"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    os_name: orm.Mapped[str]
    full_name: orm.Mapped[str]
    category: orm.Mapped[str]
    category1: orm.Mapped[str | None]
    category2: orm.Mapped[str | None]
    product_name: orm.Mapped[str]
    publisher: orm.Mapped[str]
    edition: orm.Mapped[str | None]
    market_version: orm.Mapped[str | None]
    version: orm.Mapped[str | None]
    update: orm.Mapped[str | None]
    architecture: orm.Mapped[str | None]
    lifecycle: orm.Mapped[str | None]
    taxonomy: orm.Mapped[OperatingSystemTaxonomy] = orm.relationship(
        back_populates="operating_system", uselist=False
    )
    product_url: orm.Mapped[str | None]
    product_family: orm.Mapped[str | None]
    install_date: orm.Mapped[str | None]
    release: orm.Mapped[str | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="operating_system"
    )


class Hardware(Base):
    __tablename__ = "hardware"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    full_name: orm.Mapped[str | None]
    category: orm.Mapped[str]
    category1: orm.Mapped[str]
    category2: orm.Mapped[str]
    manufacturer: orm.Mapped[str | None]
    product_name: orm.Mapped[str]
    model: orm.Mapped[str | None]
    lifecycle: orm.Mapped[str | None]
    taxonomy: orm.Mapped[HardwareTaxonomy] = orm.relationship(
        back_populates="hardware", uselist=False
    )
    product_url: orm.Mapped[str | None]
    product_family: orm.Mapped[str | None]
    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="hardware")


class UserAccountItem(Base):
    __tablename__ = "user_account_item"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    name: orm.Mapped[str] = orm.mapped_column()

    user_account_list_data_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("user_account_list_data.id")
    )
    user_account_list_data: orm.Mapped["UserAccountListData"] = orm.relationship(
        back_populates="user_account"
    )


class UserAccountListData(Base):
    __tablename__ = "user_account_list_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    user_account: orm.Mapped[list[UserAccountItem]] = orm.relationship(
        back_populates="user_account_list_data"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="user_account_list_data"
    )


class OpenPortItem(Base):
    __tablename__ = "open_port_item"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    port: orm.Mapped[int]
    description: orm.Mapped[str | None]
    protocol: orm.Mapped[str]
    detected_service: orm.Mapped[str | None]
    first_found: orm.Mapped[dt.datetime | None]
    last_updated: orm.Mapped[dt.datetime | None]

    open_port_list_data_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("open_port_list_data.id")
    )
    open_port_list_data: orm.Mapped["OpenPortListData"] = orm.relationship(
        back_populates="open_port"
    )


class OpenPortListData(Base):
    __tablename__ = "open_port_list_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    open_port: orm.Mapped[list[OpenPortItem]] = orm.relationship(
        back_populates="open_port_list_data"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="open_port_list_data"
    )


class VolumeItem(Base):
    __tablename__ = "volume_item"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    name: orm.Mapped[str]
    free: orm.Mapped[str]
    size: orm.Mapped[str]

    volume_list_data_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("volume_list_data.id")
    )
    volume_list_data: orm.Mapped["VolumeListData"] = orm.relationship(
        back_populates="volume"
    )


class VolumeListData(Base):
    __tablename__ = "volume_list_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    volume: orm.Mapped[list[VolumeItem]] = orm.relationship(
        back_populates="volume_list_data"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="volume_list_data"
    )


class NetworkInterfaceItem(Base):
    __tablename__ = "network_interface_item"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    hostname: orm.Mapped[str | None]
    address_ip_v4: orm.Mapped[ipaddress.IPv4Address | None] = orm.mapped_column(
        "address_ip_v4", sa_types.IPv4AddressType
    )
    address_ip_v6: orm.Mapped[ipaddress.IPv6Address | None] = orm.mapped_column(
        "address_ip_v6", sa_types.IPv6AddressType
    )
    mac_address: orm.Mapped[str]
    interface_name: orm.Mapped[str]
    dns_address: orm.Mapped[str | None]
    gateway_address: orm.Mapped[str]
    manufacturer: orm.Mapped[str | None]
    mac_vendor_intro_date: orm.Mapped[dt.datetime | None]
    netmask: orm.Mapped[str | None]
    addresses: orm.Mapped[str | None]

    network_interface_list_data_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("network_interface_list_data.id")
    )
    network_interface_list_data: orm.Mapped["NetworkInterfaceListData"] = (
        orm.relationship(back_populates="network_interface")
    )


class NetworkInterfaceListData(Base):
    __tablename__ = "network_interface_list_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    network_interface: orm.Mapped[list[NetworkInterfaceItem]] = orm.relationship(
        back_populates="network_interface_list_data"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="network_interface_list_data"
    )


class SoftwareItem(Base):
    __tablename__ = "software_item"

    orm_id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    id: orm.Mapped[str]
    full_name: orm.Mapped[str | None]
    software_type: orm.Mapped[str]
    is_ignored: orm.Mapped[bool]
    ignored_reason: orm.Mapped[str | None]
    category: orm.Mapped[str | None]
    category1: orm.Mapped[str | None]
    category2: orm.Mapped[str | None]
    product_name: orm.Mapped[str | None]
    component: orm.Mapped[str | None]
    publisher: orm.Mapped[str | None]
    edition: orm.Mapped[str | None]
    market_version: orm.Mapped[str | None]
    version: orm.Mapped[str | None]
    update: orm.Mapped[str | None]
    architecture: orm.Mapped[str | None]
    install_date: orm.Mapped[dt.datetime | None]
    install_path: orm.Mapped[str | None]
    last_updated: orm.Mapped[dt.datetime | None]
    last_use_date: orm.Mapped[dt.datetime | None]
    language: orm.Mapped[str | None]
    formerly_known_as: orm.Mapped[str | None]
    is_package: orm.Mapped[bool | None]
    is_package_component: orm.Mapped[bool | None]
    package_name: orm.Mapped[str | None]
    product_url: orm.Mapped[str | None]
    lifecycle: orm.Mapped[str | None]
    support_stage_desc: orm.Mapped[str | None]
    license: orm.Mapped[str | None]
    authorization: orm.Mapped[str | None]

    software_list_data_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("software_list_data.id")
    )
    software_list_data: orm.Mapped["SoftwareListData"] = orm.relationship(
        back_populates="software"
    )


class SoftwareListData(Base):
    __tablename__ = "software_list_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    software: orm.Mapped[list[SoftwareItem]] = orm.relationship(
        back_populates="software_list_data"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="software_list_data"
    )


class Activation(Base):
    __tablename__ = "activation"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    key: orm.Mapped[str]
    status: orm.Mapped[str]

    agent_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("agent.id"))
    agent: orm.Mapped["Agent"] = orm.relationship(back_populates="activations")


class Oci_Compute(Base):
    __tablename__ = "oci_compute"

    oci_id: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    tenant_id: orm.Mapped[str | None]
    tenant_name: orm.Mapped[str | None]
    compartment_id: orm.Mapped[str | None]
    compartment_name: orm.Mapped[str | None]
    image: orm.Mapped[str | None]
    shape: orm.Mapped[str | None]
    state: orm.Mapped[str]
    region: orm.Mapped[str | None]
    availability_domain: orm.Mapped[str | None]
    fault_domain: orm.Mapped[str | None]
    creation_date: orm.Mapped[dt.datetime]
    has_agent: orm.Mapped[bool]
    qualys_scanner: orm.Mapped[bool]

    oci_parent_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("oci.id"))
    oci_parent: orm.Mapped["Oci"] = orm.relationship(back_populates="compute")


class CloudTag(Base):
    __tablename__ = "cloud_tag"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    key: orm.Mapped[str]
    value: orm.Mapped[str]
    type: orm.Mapped[str]
    name_space: orm.Mapped[str]

    oci_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("oci.id"))
    oci: orm.Mapped["Oci"] = orm.relationship(back_populates="tags")


class Oci(Base):
    __tablename__ = "oci"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    compute: orm.Mapped[list[Oci_Compute]] = orm.relationship(
        "Oci_Compute", back_populates="oci_parent", uselist=False
    )
    tags: orm.Mapped[list[CloudTag]] = orm.relationship(
        "CloudTag", back_populates="oci"
    )

    cloud_provider_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("cloud_provider.id")
    )
    cloud_provider: orm.Mapped["Cloud_Provider"] = orm.relationship(
        back_populates="oci"
    )


class Cloud_Provider(Base):
    __tablename__ = "cloud_provider"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    oci: orm.Mapped[Oci] = orm.relationship(
        "Oci", back_populates="cloud_provider", uselist=False
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="cloud_provider"
    )


class Agent(Base):
    __tablename__ = "agent"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    version: orm.Mapped[str | None]
    configuration_profile: orm.Mapped[str | None]
    activations: orm.Mapped[list[Activation]] = orm.relationship(
        "Activation", back_populates="agent"
    )
    connected_from: orm.Mapped[str | None]
    last_activity: orm.Mapped[dt.datetime]
    last_checked_in: orm.Mapped[dt.datetime]
    last_inventory: orm.Mapped[dt.datetime]
    udc_manifest_assigned: orm.Mapped[bool | None]
    error_status: orm.Mapped[bool | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="agent")


class Sensor(Base):
    __tablename__ = "sensor"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    activated_for_modules: orm.Mapped[list[str]] = orm.mapped_column(
        sa_pg.ARRAY(sa.String)
    )
    pending_activation_for_modules: orm.Mapped[list[str] | None] = orm.mapped_column(
        sa_pg.ARRAY(sa.String)
    )
    last_v_m_scan: orm.Mapped[dt.datetime]
    last_compliance_scan: orm.Mapped[dt.datetime]
    last_full_scan: orm.Mapped[dt.datetime]
    last_vm_scan_date_scanner: orm.Mapped[dt.datetime]
    last_vm_scan_date_agent: orm.Mapped[dt.datetime]
    last_pc_scan_date_scanner: orm.Mapped[dt.datetime]
    last_pc_scan_date_agent: orm.Mapped[dt.datetime]
    first_easm_scan_date: orm.Mapped[dt.datetime | None]
    last_easm_scan_date: orm.Mapped[dt.datetime | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="sensor")


class Inventory(Base):
    __tablename__ = "inventory"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    source: orm.Mapped[str]
    created: orm.Mapped[dt.datetime]
    last_updated: orm.Mapped[dt.datetime]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="inventory")


class Container(Base):
    __tablename__ = "container"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    product: orm.Mapped[str | None]
    version: orm.Mapped[str | None]
    no_of_containers: orm.Mapped[int]
    no_of_images: orm.Mapped[int]
    has_sensor: orm.Mapped[bool | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="container")


class Activity(Base):
    __tablename__ = "activity"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    source: orm.Mapped[str]
    last_scanned_date: orm.Mapped[dt.datetime]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="activity")


taglist_tagitem_association_table = sa.Table(
    "taglist_tagitem",
    Base.metadata,
    sa.Column("taglist_id", sa.Integer, sa.ForeignKey("tag_list.id"), primary_key=True),
    sa.Column(
        "tagitem_id", sa.Integer, sa.ForeignKey("tag_item.tag_id"), primary_key=True
    ),
)


class TagItem(Base):
    __tablename__ = "tag_item"

    tag_id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    tag_name: orm.Mapped[str]
    foreground_color: orm.Mapped[int]
    background_color: orm.Mapped[int]
    business_impact: orm.Mapped[str | None]
    criticality_score: orm.Mapped[int | None]

    tag_list: orm.Mapped[list["TagList"]] = orm.relationship(
        secondary=taglist_tagitem_association_table, back_populates="tag"
    )


class TagList(Base):
    __tablename__ = "tag_list"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    tag: orm.Mapped[list[TagItem]] = orm.relationship(
        secondary=taglist_tagitem_association_table, back_populates="tag_list"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="tag_list")


class ServiceItem(Base):
    __tablename__ = "service_item"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    description: orm.Mapped[str | None]
    name: orm.Mapped[str]
    status: orm.Mapped[str]

    service_list_id = orm.mapped_column(sa.ForeignKey("service_list.id"))
    service_list: orm.Mapped["ServiceList"] = orm.relationship(
        "ServiceList", back_populates="service"
    )


class ServiceList(Base):
    __tablename__ = "service_list"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    service: orm.Mapped[list[ServiceItem]] = orm.relationship(
        "ServiceItem", back_populates="service_list"
    )

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="service_list"
    )


class LastLocation(Base):
    __tablename__ = "last_location"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    city: orm.Mapped[str | None]
    state: orm.Mapped[str | None]
    country: orm.Mapped[str]
    name: orm.Mapped[str]
    continent: orm.Mapped[str]
    postal: orm.Mapped[str | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(
        back_populates="last_location"
    )


class Criticality(Base):
    __tablename__ = "criticality"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    score: orm.Mapped[int]
    is_default: orm.Mapped[bool]
    last_updated: orm.Mapped[dt.datetime | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="criticality")


class Processor(Base):
    __tablename__ = "processor"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    description: orm.Mapped[str | None]
    speed: orm.Mapped[int | None]
    numCPUs: orm.Mapped[int | None]
    no_of_socket: orm.Mapped[int | None]
    threads_per_core: orm.Mapped[int | None]
    cores_per_socket: orm.Mapped[int | None]
    multithreading_status: orm.Mapped[str | None]

    asset_item_id: orm.Mapped[int] = orm.mapped_column(
        sa.ForeignKey("asset_item.asset_id")
    )
    asset_item: orm.Mapped["AssetItem"] = orm.relationship(back_populates="processor")


class AssetItem(Base):
    __tablename__ = "asset_item"

    asset_id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    assetUUID: orm.Mapped[str]
    host_id: orm.Mapped[int | None]
    last_modified_date: orm.Mapped[dt.datetime]
    agent_id: orm.Mapped[str | None]
    created_date: orm.Mapped[dt.datetime]
    sensor_last_updated_date: orm.Mapped[dt.datetime]
    asset_type: orm.Mapped[str | None]
    address: orm.Mapped[ipaddress.IPv4Address | ipaddress.IPv6Address] = (
        orm.mapped_column("address", sa_types.IPAddressGenericType)
    )
    dns_name: orm.Mapped[str | None]
    asset_name: orm.Mapped[str]
    netbios_name: orm.Mapped[str | None]
    time_zone: orm.Mapped[str | None]
    bios_description: orm.Mapped[str | None]
    last_boot: orm.Mapped[dt.datetime | None]
    total_memory: orm.Mapped[int | None]
    cpu_count: orm.Mapped[int | None]
    last_logged_on_user: orm.Mapped[str | None]
    domain_role: orm.Mapped[str | None]
    hwUUID: orm.Mapped[str | None]
    bios_serial_number: orm.Mapped[str | None]
    bios_asset_tag: orm.Mapped[str | None]
    is_container_host: orm.Mapped[bool | None]
    operating_system: orm.Mapped[OperatingSystem] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    hardware: orm.Mapped[Hardware] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    user_account_list_data: orm.Mapped[UserAccountListData] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    open_port_list_data: orm.Mapped[OpenPortListData] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    volume_list_data: orm.Mapped[VolumeListData] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    network_interface_list_data: orm.Mapped[NetworkInterfaceListData | None] = (
        orm.relationship(back_populates="asset_item", uselist=False)
    )
    software_list_data: orm.Mapped[SoftwareListData] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    provider: orm.Mapped[str | None]
    cloud_provider: orm.Mapped[Cloud_Provider | None] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    agent: orm.Mapped[Agent | None] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    sensor: orm.Mapped[Sensor] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    container: orm.Mapped[Container | None] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    inventory: orm.Mapped[Inventory] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    activity: orm.Mapped[Activity] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    tag_list: orm.Mapped[TagList | None] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    service_list: orm.Mapped[ServiceList] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    last_location: orm.Mapped[LastLocation | None] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    criticality: orm.Mapped[Criticality] = orm.relationship(
        back_populates="asset_item", uselist=False
    )
    business_information: orm.Mapped[str | None]
    assigned_location: orm.Mapped[str | None]
    business_app_list_data: orm.Mapped[str | None]
    risk_score: orm.Mapped[int | None]
    passive_sensor: orm.Mapped[str | None]
    domain: orm.Mapped[list[str] | None] = orm.mapped_column(sa_pg.ARRAY(sa.String))
    subdomain: orm.Mapped[list[str] | None] = orm.mapped_column(sa_pg.ARRAY(sa.String))
    missing_software: orm.Mapped[str | None]
    whois: orm.Mapped[list[str] | None] = orm.mapped_column(sa_pg.ARRAY(sa.String))
    isp: orm.Mapped[str | None]
    asn: orm.Mapped[str | None]
    easm_tags: orm.Mapped[str | None]
    hosting_category1: orm.Mapped[str | None]
    custom_attributes: orm.Mapped[str | None]
    processor: orm.Mapped[Processor] = orm.relationship(
        back_populates="asset_item", uselist=False
    )



"""
Qualys VMDR API Module

Typical usage example:
vmdr_orm = vmdr.HostListDetectionORM()
vmdr_orm.load()
session = orm.Session(vmdr_orm.engine):
host_list = session.query(HostORM).all()
stmt = session.query(HostORM)
hosts = vmdr_orm.query(stmt)[0]
host = hosts.host[0]
session.close()
"""

import datetime
import ipaddress
import re
from typing import Any, Literal

import sqlalchemy.orm as orm
from lxml import etree
from lxml.etree import XMLSyntaxError
from psycopg import OperationalError as pgOperationalError
from sqlalchemy.exc import OperationalError as saOperationalError

from . import URLS, qutils
from .base import QualysAPIBase, QualysORMMixin
from .models.vmdr import (
    asset_group_list_output,
    host_list_orm,
    host_list_output,
    host_list_vm_detection_orm,
    host_list_vm_detection_output,
    knowledgebase_orm,
    knowledgebase_output,
    map_report,
    map_report_list,
    scan_list_output,
    simple_return,
)


class VmdrAPI(QualysAPIBase):
    """Qualys VMDR API Class.  Contains methods for interacting with the VMDR API."""

    def host_list(
        self,
        *,
        show_asset_id: bool | None = None,
        details: Literal["Basic", "Basic/AGs", "All", "All/AGs", "None"] | None = None,
        os_pattern: str | None = None,
        truncation_limit: int | None = None,
        ips: (
            list[str | ipaddress.IPv4Address | ipaddress.IPv6Address]
            | str
            | ipaddress.IPv4Address
            | ipaddress.IPv6Address
            | None
        ) = None,
        ag_ids: int | list[int] | None = None,
        ag_titles: str | list[str] | None = None,
        ids: int | list[int] | None = None,
        id_min: int | None = None,
        id_max: int | None = None,
        network_ids: int | list[int] | None = None,
        compliance_enabled: bool | None = None,
        no_vm_scan_since: datetime.datetime | None = None,
        no_compliance_scan_since: datetime.datetime | None = None,
        vm_scan_since: datetime.datetime | None = None,
        compliance_scan_since: datetime.datetime | None = None,
        vm_processed_before: datetime.datetime | None = None,
        vm_processed_after: datetime.datetime | None = None,
        vm_scan_date_before: datetime.datetime | None = None,
        vm_scan_date_after: datetime.datetime | None = None,
        vm_auth_scan_date_before: datetime.datetime | None = None,
        vm_auth_scan_date_after: datetime.datetime | None = None,
        scap_scan_since: datetime.datetime | None = None,
        no_scap_scan_since: datetime.datetime | None = None,
        use_tags: bool | None = None,
        tag_set_by: Literal["id", "name"] | None = None,
        tag_include_selector: Literal["any", "all"] | None = None,
        tag_exclude_selector: Literal["any", "all"] | None = None,
        tag_set_include: list[str | int] | None = None,
        tag_set_exclude: list[str | int] | None = None,
        show_tags: bool | None = None,
        show_trurisk: bool | None = None,
        trurisk_min: int | None = None,
        trurisk_max: int | None = None,
        show_trurisk_factors: bool | None = None,
    ) -> tuple[list[host_list_output.Host], bool, int]:
        """Get a list of hosts from the VMDR API.  A value of None for the parameters will use their
            default values in the API.

        Args:
            ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.

        Returns:
            tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                host_list_vm_detection_output.HostList object, a boolean indicating whether the
                results were truncated, and the next id_min to use for the next call.
        """

        if ips is not None:
            if not isinstance(ips, list):
                ips = [ips]
            ipv4_ips: (
                list[str | ipaddress.IPv4Address | ipaddress.IPv6Address] | None
            ) = [
                ip
                for ip in ips
                if isinstance(ip, ipaddress.IPv4Address)
                or isinstance(ipaddress.ip_address(ip), ipaddress.IPv4Address)
            ]
            ipv6_ips: (
                list[str | ipaddress.IPv4Address | ipaddress.IPv6Address] | None
            ) = [
                ip
                for ip in ips
                if isinstance(ip, ipaddress.IPv6Address)
                or isinstance(ipaddress.ip_address(ip), ipaddress.IPv6Address)
            ]

            if not ipv4_ips:
                ipv4_ips = None
            if not ipv6_ips:
                ipv6_ips = None
        else:
            ipv4_ips = None
            ipv6_ips = None

        # Cannot pass both IPv4 and IPv6 addresses in the same call.
        if ipv4_ips and ipv6_ips:
            raise ValueError(
                "Cannot pass both IPv4 and IPv6 addresses in the same call."
            )

        params: dict[str, Any] = {
            "show_asset_id": show_asset_id,
            "details": details,
            "os_pattern": os_pattern,
            "truncation_limit": truncation_limit,
            "ips": ipv4_ips,
            "ipv6": ipv6_ips,
            "ag_ids": ag_ids,
            "ag_titles": ag_titles,
            "ids": ids,
            "id_min": id_min,
            "id_max": id_max,
            "network_ids": network_ids,
            "compliance_enabled": compliance_enabled,
            "no_vm_scan_since": no_vm_scan_since,
            "no_compliance_scan_since": no_compliance_scan_since,
            "vm_scan_since": vm_scan_since,
            "compliance_scan_since": compliance_scan_since,
            "vm_processed_before": vm_processed_before,
            "vm_processed_after": vm_processed_after,
            "vm_scan_date_before": vm_scan_date_before,
            "vm_scan_date_after": vm_scan_date_after,
            "vm_auth_scan_date_before": vm_auth_scan_date_before,
            "vm_auth_scan_date_after": vm_auth_scan_date_after,
            "scap_scan_since": scap_scan_since,
            "no_scap_scan_since": no_scap_scan_since,
            "use_tags": use_tags,
            "tag_set_by": tag_set_by,
            "tag_include_selector": tag_include_selector,
            "tag_exclude_selector": tag_exclude_selector,
            "tag_set_include": tag_set_include,
            "tag_set_exclude": tag_set_exclude,
            "show_tags": show_tags,
            "show_trurisk": show_trurisk,
            "trurisk_min": trurisk_min,
            "trurisk_max": trurisk_max,
            "show_trurisk_factors": show_trurisk_factors,
        }
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.host_list, params=params_cleaned).text
        match = re.search(
            r"<HOST_LIST_OUTPUT>.*?</HOST_LIST_OUTPUT>", raw_response, re.DOTALL
        )
        if match is None:
            raise ValueError("Cannot find HOST_LIST_OUTPUT in response.")
        host_list_output_str = match.group(0)
        host_list_output_obj = host_list_output.HostListOutput.from_xml(
            host_list_output_str
        )
        if host_list_output_obj.response.host_list is None:
            raise ValueError("Response has no host_list")

        host_list = host_list_output_obj.response.host_list
        warning = host_list_output_obj.response.warning
        if warning is None:
            return host_list, False, 0
        next_id_match = re.search(r"id_min=(\d+)", warning.url)
        if next_id_match is None:
            raise ValueError(
                "Unable to parse URL in warning message. No id_min found.\n"
                f"{warning.url}"
            )
        truncated = True
        next_id_min = int(next_id_match.group(1))

        return host_list, truncated, next_id_min

    def host_list_vm_detection(
        self,
        *,
        ids: int | list[int] | None = None,
        truncation_limit: int | None = None,
        id_min: int | None = None,
        qids: int | list[int] | None = None,
        show_qds: bool | None = None,
        qds_min: int | None = None,
        qds_max: int | None = None,
        arf_kernel_filter: int | None = None,
        show_igs: bool | None = None,
    ) -> tuple[list[host_list_vm_detection_output.Host], bool, int]:
        """Get a list of hosts with associated vulnerability detections from the VMDR API.  A
        value of None for the parameters will use their default values in the API.

            Args:
                ids (int | list[int] | None, optional): Host IDs to query. Defaults to None.
                truncation_limit (int | None, optional): Maximum number of hosts to return. Defaults
                     to None.
                id_min (int | None, optional): Minimum host list ID to return. Defaults to None.

            Returns:
                tuple[host_list_vm_detection_output.HostList, bool, int]: A tuple containing the
                    host_list_vm_detection_output.HostList object, a boolean indicating whether the
                    results were truncated, and the next id_min to use for the next call.
        """

        params = {
            "ids": ids,
            "truncation_limit": truncation_limit,
            "id_min": id_min,
            "qids": qids,
            "show_qds": show_qds,
            "qds_min": qds_min,
            "qds_max": qds_max,
            "arf_kernel_filter": arf_kernel_filter,
            "show_igs": show_igs,
        }
        cleaned_params = qutils.clean_dict(params)
        cleaned_params["action"] = "list"

        # Detections results can be quite large, so we need to set the parser to allow for large
        # trees.
        parser = etree.XMLParser(huge_tree=True)

        raw_response = self.get(URLS.host_list_vm_detection, params=cleaned_params).text
        match = re.search(
            r"<HOST_LIST_VM_DETECTION_OUTPUT>.*?</HOST_LIST_VM_DETECTION_OUTPUT>",
            raw_response,
            re.DOTALL,
        )
        if match is None:
            raise ValueError("Cannot find HOST_LIST_VM_DETECTION_OUTPUT in response.")
        host_list_vm_detection_output_str = match.group(0)
        host_list_vm_detection_output_obj = (
            host_list_vm_detection_output.HostListVMDetectionOutput.from_xml(
                host_list_vm_detection_output_str, parser=parser
            )
        )
        if host_list_vm_detection_output_obj.response.host_list is None:
            raise ValueError("Response has no host_list")

        host_list = host_list_vm_detection_output_obj.response.host_list
        warning = host_list_vm_detection_output_obj.response.warning
        if warning is None:
            return host_list, False, 0
        next_id_match = re.search(r"id_min=(\d+)", warning.url)
        if next_id_match is None:
            raise ValueError(
                "Unable to parse URL in warning message. No id_min found.\n"
                f"{warning.url}"
            )
        truncated = True
        next_id_min = int(next_id_match.group(1))

        return host_list, truncated, next_id_min

    def knowledgebase(
        self,
        *,
        details: str | None = None,
        ids: int | list[int] | None = None,
        id_min: int | None = None,
        id_max: int | None = None,
    ) -> list[knowledgebase_output.Vuln]:
        """Get a list of vulnerabilities from the VMDR API.  A value of None for the parameters
        will use their default values in the API.

            Args:
                details (str | None, optional): Details to return. Defaults to None.
                ids (int | list[int] | None, optional): Vulnerability IDs to query. Defaults to
                  None.

            Returns:
                knowledge_base_vuln_list_output.VulnList: A VulnList object containing the list of
                    vulnerabilities.
        """
        params = {"details": details, "ids": ids, "id_min": id_min, "id_max": id_max}
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.knowledgebase, params=params_cleaned).text
        match = re.search(
            r"<KNOWLEDGE_BASE_VULN_LIST_OUTPUT>.*?</KNOWLEDGE_BASE_VULN_LIST_OUTPUT>",
            raw_response,
            re.DOTALL,
        )
        if match is None:
            raise ValueError("Cannot find KNOWLEDGE_BASE_VULN_LIST_OUTPUT in response.")
        knowledge_base_output_str = match.group(0)
        try:
            knowledge_base_output_obj = (
                knowledgebase_output.KnowledgeBaseOutput.from_xml(
                    knowledge_base_output_str
                )
            )
        except XMLSyntaxError as e:
            self.log.error(
                f"Error parsing XML response: {e}\nResponse:\n{knowledge_base_output_str}"
            )
        if knowledge_base_output_obj.response.vuln_list is None:
            raise ValueError("Response has no vuln_list")

        return knowledge_base_output_obj.response.vuln_list

    def launch_vm_scan(
        self,
        *,
        scan_title: str | None = None,
        iscanner_name: str | None = None,
        option_title: str | None = None,
        fqdn: str | list[str] | None = None,
    ) -> simple_return.SimpleReturn:
        params = {
            "scan_title": scan_title,
            "iscanner_name": iscanner_name,
            "option_title": option_title,
            "fqdn": fqdn,
        }
        params["action"] = "launch"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.post(
            URLS.launch_vm_scan, params=params_cleaned
        ).text.encode("utf-8")
        launch_vm_scan_output_obj = simple_return.SimpleReturn.from_xml(raw_response)

        if launch_vm_scan_output_obj.response.text != "New vm scan launched":
            raise ValueError(
                f"Failed to launch VM scan.  Response: {launch_vm_scan_output_obj.response.text}"
            )

        return launch_vm_scan_output_obj

    def vm_scan_list(
        self, *, scan_ref: str | None = None
    ) -> scan_list_output.ScanListOutput:
        params = {"scan_ref": scan_ref}
        params["action"] = "list"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.vm_scan_list, params=params_cleaned).text.encode(
            "utf-8"
        )
        vm_scan_list_output_obj = scan_list_output.ScanListOutput.from_xml(raw_response)

        return vm_scan_list_output_obj

    def map_report_list(
        self, *, last: bool | None = None, domain: str | None = None
    ) -> map_report_list.MapReportList:
        params = {"domain": domain}
        if last:
            params["last"] = "yes"
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(
            URLS.map_report_list, params=params_cleaned
        ).text.encode("utf-8")
        map_report_list_output_obj = map_report_list.MapReportList.from_xml(
            raw_response
        )

        return map_report_list_output_obj

    def download_saved_map_report(self, *, ref: str | None) -> map_report.Map:
        params = {"ref": ref}
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(
            URLS.download_saved_map_report, params=params_cleaned
        ).text.encode("utf-8")
        map_report_output_obj = map_report.Map.from_xml(raw_response)

        return map_report_output_obj

    def asset_group_list(self) -> list[asset_group_list_output.AssetGroup]:
        params = {"action": "list"}
        params_cleaned = qutils.clean_dict(params)

        raw_response = self.get(URLS.asset_group, params=params_cleaned).text.encode(
            "utf-8"
        )
        asset_group_list_output_obj = (
            asset_group_list_output.AssetGroupListOutput.from_xml(raw_response)
        )
        if asset_group_list_output_obj.response.asset_group_list is None:
            raise ValueError("Response has no asset_group_list")

        return asset_group_list_output_obj.response.asset_group_list

    def edit_asset_group(
        self,
        *,
        id: int,
        set_ips: list[str | ipaddress.IPv4Address | ipaddress.IPv4Network]
        | None = None,
    ) -> simple_return.SimpleReturn:
        params = {"action": "edit"}
        data = {"id": str(id)}
        if set_ips is not None:
            ips_str = ",".join([str(ip) for ip in set_ips])
            data["set_ips"] = ips_str

        raw_response = self.post(
            URLS.asset_group,
            params=params,
            data=data,
            content_type="application/x-www-form-urlencoded",
        ).text.encode("utf-8")
        edit_asset_group_output_obj = simple_return.SimpleReturn.from_xml(raw_response)

        if (
            edit_asset_group_output_obj.response.text
            != "Asset Group Updated Successfully"
        ):
            raise ValueError(
                f"Failed to edit asset group.  Response: {edit_asset_group_output_obj.response.text}"
            )
        return edit_asset_group_output_obj


class HostListORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List ORM Class.  Contains methods for loading hosts into an ORM database."""

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            load_func (Any): Function to call to get hosts.
            **kwargs (Any): Keyword arguments to pass to host_list.
        """

        def load_set(to_load: list[host_list_orm.Host]) -> None:
            """Load a single set of hosts into the ORM database.

            Args:
                to_load (list[host_list_orm.Host]): List of hosts to load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                for obj in to_load:
                    session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 10000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            hosts, truncated, next_id_min = self.host_list(**kwargs)
            to_load = [qutils.to_orm_object(host, host_list_orm.Host) for host in hosts]
            load_set(to_load)


class HostListVMDetectionORM(VmdrAPI, QualysORMMixin):
    """Qualys VMDR Host List Detection ORM Class.  Contains methods for loading host
    detections into an ORM database.
    """

    def __init__(self, echo: bool = False) -> None:
        """Initialize the Host List Detection ORM Class.

        Args:
            echo (bool, optional): Whether to echo SQL statements. Defaults to False.
        """
        VmdrAPI.__init__(self)
        self.orm_base = host_list_vm_detection_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        """Load hosts into the ORM database.

        Args:
            load_func (Any): Function to call to get hosts.
            **kwargs (Any): Keyword arguments to pass to host_list_vm_detection.
        """

        def load_set(to_load: list[host_list_vm_detection_orm.Host]) -> None:
            """Load a single set of hosts into the ORM database.

            Args:
                to_load (list[host_list_vm_detection_orm.Host]): List of hosts to load.
            """
            with orm.Session(self.engine) as session:
                session.add_all(to_load)
                # for obj in to_load:
                #     session.merge(obj)
                session.commit()

        kwargs.setdefault("truncation_limit", 1000)
        truncated = True
        next_id_min = None
        while truncated:
            kwargs["id_min"] = next_id_min
            success = False
            # Attempt to fetch hosts, halving the truncation limit on errors
            # which typically means the system ran out of memory, presumably due to the size
            # of the input.
            while not success:
                try:
                    hosts, truncated, next_id_min = self.host_list_vm_detection(
                        **kwargs
                    )
                    to_load = qutils.to_orm_objects(
                        hosts, host_list_vm_detection_orm.Host
                    )
                    load_set(to_load)
                    success = True
                except (
                    XMLSyntaxError,
                    OverflowError,
                    pgOperationalError,
                    saOperationalError,
                ):
                    if kwargs["truncation_limit"] > 1:
                        kwargs["truncation_limit"] //= 2
                    else:
                        raise
            kwargs["truncation_limit"] = 1000
            # to_load = [
            #     qutils.to_orm_object(host, host_list_vm_detection_orm.Host)
            #     for host in hosts
            # ]


class KnowledgebaseORM(VmdrAPI, QualysORMMixin):
    def __init__(self, echo: bool = False) -> None:
        VmdrAPI.__init__(self)
        self.orm_base = knowledgebase_orm.Base  # type: ignore
        QualysORMMixin.__init__(self, self, echo=echo)

    def load(self, **kwargs: Any) -> None:
        MAX_QID = 10000000
        NUM_PER_CALL = 50000
        for id_min in range(1, MAX_QID, NUM_PER_CALL):
            kwargs["id_min"] = id_min
            kwargs["id_max"] = id_min + (NUM_PER_CALL - 1)
            vulns = self.knowledgebase(**kwargs)
            to_load = qutils.to_orm_objects(vulns, knowledgebase_orm.Vuln)
            with orm.Session(self.engine) as session:
                for item in to_load:
                    session.merge(item)
                session.commit()



import datetime

from pydantic_xml import BaseXmlModel, element, wrapped


class OptionProfile(BaseXmlModel):
    title: str = element(tag="TITLE")
    default_flag: bool | None = element(tag="DEFAULT_FLAG", default=None)


class Status(BaseXmlModel):
    state: str = element(tag="STATE")
    sub_state: str | None = element(tag="SUB_STATE", default=None)


class Client(BaseXmlModel):
    id: int = element(tag="ID")
    name: str = element(tag="NAME")


class Scan(BaseXmlModel):
    id: int | None = element(tag="ID", default=None)
    ref: str = element(tag="REF")
    scan_type: str | None = element(tag="SCAN_TYPE", default=None)
    type: str = element(tag="TYPE")
    title: str = element(tag="TITLE")
    client: Client | None = element(tag="CLIENT", default=None)
    user_login: str = element(tag="USER_LOGIN")
    launch_datetime: datetime.datetime = element(tag="LAUNCH_DATETIME")
    duration: str = element(tag="DURATION")
    processing_priority: str | None = element(tag="PROCESSING_PRIORITY", default=None)
    processed: bool = element(tag="PROCESSED")
    status: Status | None = element(tag="STATUS", default=None)
    target: str | None = element(tag="TARGET", default=None)
    asset_group_title_list: list[str] | None = wrapped(
        "ASSET_GROUP_TITLE_LIST", element(tag="ASSET_GROUP_TITLE", default=None)
    )
    option_profile: OptionProfile | None = element(tag="OPTION_PROFILE", default=None)


class Response(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    scan_list: list[Scan] | None = wrapped(
        "SCAN_LIST", element(tag="SCAN", default=None)
    )


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] | None = wrapped(
        "PARAM_LIST", element(tag="PARAM", default=None)
    )
    post_data: str | None = element(tag="POST_DATA")


class ScanListOutput(BaseXmlModel, tag="SCAN_LIST_OUTPUT"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE")



"""ORM data model for host_list_vm_detection"""

import datetime as dt
import ipaddress

import sqlalchemy as sa
import sqlalchemy.orm as orm

from .. import sa_types


class Base(orm.DeclarativeBase):
    metadata = sa.MetaData(schema="host_list_vm_detection")


class Attribute(Base):
    __tablename__ = "attribute"

    name: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    last_status: orm.Mapped[str | None]
    value: orm.Mapped[str | None]
    last_success_date: orm.Mapped[dt.datetime | None]
    last_error_date: orm.Mapped[dt.datetime | None]
    last_error: orm.Mapped[str | None]

    azure_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("azure.id"))
    azure: orm.Mapped["Azure | None"] = orm.relationship(back_populates="attribute")
    google_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("google.id"))
    google: orm.Mapped["Google | None"] = orm.relationship(back_populates="attribute")
    ec2_id: orm.Mapped[int | None] = orm.mapped_column(sa.ForeignKey("ec2.id"))
    ec2: orm.Mapped["Ec2 | None"] = orm.relationship(back_populates="attribute")


class CloudTag(Base):
    __tablename__ = "cloud_tag"

    name: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    value: orm.Mapped[str | None]
    last_success_date: orm.Mapped[dt.datetime | None]

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="cloud_provider_tags")


class DnsData(Base):
    __tablename__ = "dns_data"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    hostname: orm.Mapped[str | None]
    domain: orm.Mapped[str | None]
    fqdn: orm.Mapped[str | None]

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="dns_data")


class Qds(Base):
    __tablename__ = "qds"
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ["detection_unqiue_vuln_id", "detection_qid"],
            ["detection.unique_vuln_id", "detection.qid"],
        ),
    )

    severity: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    value: orm.Mapped[int] = orm.mapped_column(primary_key=True)

    detection_unqiue_vuln_id: orm.Mapped[int]
    detection_qid: orm.Mapped[int]
    detection: orm.Mapped["Detection"] = orm.relationship(back_populates="qds")


class QdsFactor(Base):
    __tablename__ = "qds_factor"

    name: orm.Mapped[str] = orm.mapped_column(primary_key=True)
    value: orm.Mapped[int] = orm.mapped_column(primary_key=True)

    qds_factors_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("qds_factors.id"))
    qds_factors: orm.Mapped["QdsFactors"] = orm.relationship(
        back_populates="qds_factor"
    )


class Tag(Base):
    __tablename__ = "tag"

    tag_id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    name: orm.Mapped[str]
    color: orm.Mapped[str | None]
    background_color: orm.Mapped[str | None]

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="tags")


class Azure(Base):
    __tablename__ = "azure"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="azure")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata_.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="azure")


class Ec2(Base):
    __tablename__ = "ec2"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="ec2")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata_.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="ec2")


class Google(Base):
    __tablename__ = "google"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    attribute: orm.Mapped[list[Attribute]] = orm.relationship(back_populates="google")

    metadata_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("metadata_.id"))
    metadata_: orm.Mapped["Metadata"] = orm.relationship(back_populates="google")


class QdsFactors(Base):
    __tablename__ = "qds_factors"
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ["detection_unqiue_vuln_id", "detection_qid"],
            ["detection.unique_vuln_id", "detection.qid"],
        ),
    )

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    qds_factor: orm.Mapped[list[QdsFactor]] = orm.relationship(
        back_populates="qds_factors"
    )

    detection_unqiue_vuln_id: orm.Mapped[int]
    detection_qid: orm.Mapped[int]
    detection: orm.Mapped["Detection"] = orm.relationship(back_populates="qds_factors")


class Detection(Base):
    __tablename__ = "detection"

    unique_vuln_id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    qid: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    type: orm.Mapped[str | None]
    severity: orm.Mapped[int | None]
    port: orm.Mapped[int | None]
    protocol: orm.Mapped[str | None]
    fqdn: orm.Mapped[str | None]
    ssl: orm.Mapped[bool | None]
    instance: orm.Mapped[str | None]
    results: orm.Mapped[str | None]
    status: orm.Mapped[str | None]
    first_found_datetime: orm.Mapped[dt.datetime | None]
    last_found_datetime: orm.Mapped[dt.datetime | None]
    source: orm.Mapped[str | None]
    qds: orm.Mapped[Qds | None] = orm.relationship(
        back_populates="detection", uselist=False
    )
    qds_factors: orm.Mapped[list[QdsFactors]] = orm.relationship(
        back_populates="detection", uselist=True
    )
    times_found: orm.Mapped[int | None]
    last_test_datetime: orm.Mapped[dt.datetime | None]
    last_update_datetime: orm.Mapped[dt.datetime | None]
    last_fixed_datetime: orm.Mapped[dt.datetime | None]
    first_reopened_datetime: orm.Mapped[dt.datetime | None]
    last_reopened_datetime: orm.Mapped[dt.datetime | None]
    times_reopened: orm.Mapped[int | None]
    service: orm.Mapped[str | None]
    is_ignored: orm.Mapped[bool | None]
    is_disabled: orm.Mapped[bool | None]
    affect_running_kernel: orm.Mapped[bool | None]
    affect_running_service: orm.Mapped[bool | None]
    affect_exploitable_config: orm.Mapped[bool | None]
    last_processed_datetime: orm.Mapped[dt.datetime | None]
    asset_cve: orm.Mapped[str | None]

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="detections")


class Metadata(Base):
    __tablename__ = "metadata_"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True, autoincrement=True)
    ec2: orm.Mapped[list[Ec2]] = orm.relationship(back_populates="metadata_")
    google: orm.Mapped[list[Google]] = orm.relationship(back_populates="metadata_")
    azure: orm.Mapped[list[Azure]] = orm.relationship(back_populates="metadata_")

    host_id: orm.Mapped[int] = orm.mapped_column(sa.ForeignKey("host.id"))
    host: orm.Mapped["Host"] = orm.relationship(back_populates="metadata_")


class Host(Base):
    __tablename__ = "host"

    id: orm.Mapped[int] = orm.mapped_column(primary_key=True)
    asset_id: orm.Mapped[int | None]
    ip: orm.Mapped[ipaddress.IPv4Address | None] = orm.mapped_column(
        "ip", sa_types.IPv4AddressType
    )
    ipv6: orm.Mapped[ipaddress.IPv6Address | None] = orm.mapped_column(
        "ipv6", sa_types.IPv6AddressType
    )
    tracking_method: orm.Mapped[str | None]
    network_id: orm.Mapped[int | None]
    os: orm.Mapped[str | None]
    os_cpe: orm.Mapped[str | None]
    dns: orm.Mapped[str | None]
    dns_data: orm.Mapped[DnsData | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    cloud_provider: orm.Mapped[str | None]
    cloud_service: orm.Mapped[str | None]
    cloud_resource_id: orm.Mapped[str | None]
    ec2_instance_id: orm.Mapped[str | None]
    netbios: orm.Mapped[str | None]
    qg_hostid: orm.Mapped[str | None]
    last_scan_datetime: orm.Mapped[dt.datetime | None]
    last_vm_scanned_date: orm.Mapped[dt.datetime | None]
    last_vm_scanned_duration: orm.Mapped[int | None]
    last_vm_auth_scanned_date: orm.Mapped[dt.datetime | None]
    last_vm_auth_scanned_duration: orm.Mapped[int | None]
    last_pc_scanned_date: orm.Mapped[dt.datetime | None]
    tags: orm.Mapped[list[Tag]] = orm.relationship(back_populates="host", uselist=True)
    metadata_: orm.Mapped[Metadata | None] = orm.relationship(
        back_populates="host", uselist=False
    )
    cloud_provider_tags: orm.Mapped[list[CloudTag]] = orm.relationship(
        back_populates="host", uselist=True
    )
    detections: orm.Mapped[list[Detection]] = orm.relationship(
        back_populates="host", uselist=True
    )



import dataclasses
from typing import Any

from . import URLS
from .base import QualysAPIBase
from .models.certview import instances_output


@dataclasses.dataclass
class Filter:
    field: str
    value: str
    operator: str

    def to_dict(self) -> dict[str, Any]:
        """Convert the filter to a dictionary.

        Returns:
            dict[str, Any]: A dictionary representation of the filter.
        """
        return dataclasses.asdict(self)


@dataclasses.dataclass
class FilterRequest:
    filters: list[Filter]
    operation: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert the filter request to a dictionary.

        Returns:
            dict[str, Any]: A dictionary representation of the filter request.
        """

        if len(self.filters) > 1 and self.operation is None:
            raise ValueError(
                "Operation must be specified when multiple filters are provided."
            )
        ret: dict[str, Any] = {
            "filters": [filter_.to_dict() for filter_ in self.filters]
        }
        if self.operation is not None:
            ret["operation"] = self.operation
        return ret


class CertViewAPI(QualysAPIBase):
    """Qualys CertView API Class.  Contains methods for interacting with the CertView API."""

    def list_instances(
        self,
        *,
        filter_request: FilterRequest | None = None,
        page_number: int | None = None,
        page_size: int | None = None,
    ) -> list[instances_output.Instance]:
        """List instances in CertView.

        Args:
            filter (Filter | None): Optional filter to apply to the list of instances.
            page_number (int | None): Optional page number for pagination.
            page_size (int | None): Optional page size for pagination.

        Returns:
            list[instances_output.Instance]: A list of instances in CertView.
        """

        data: dict[str, str | dict[str, Any]] = {}
        if filter_request is not None:
            data["filterRequest"] = filter_request.to_dict()
        if page_number is not None:
            data["pageNumber"] = str(page_number)
        if page_size is not None:
            data["pageSize"] = str(page_size)

        response = self.post(
            URLS.list_instances,
            json=data,
            content_type="application/json",
            accept="application/json",
        )
        response.raise_for_status()
        response_json = response.json()
        instances = [
            instances_output.Instance(**instance) for instance in response_json
        ]
        return instances

    def add_bulk_external_sites(self, *, sites: list[str]) -> None:
        """Add a list of external sites to CertView.

        Args:
            sites (list[str]): A list of external sites to add to CertView.
        """

        # A single CSV file can contain up to 1000 records, so add in batches of 1000
        for i in range(0, len(sites), 1000):
            sites_batch = sites[i : i + 1000]
            csv_string = "Sites\n" + "\n".join(sites_batch) + "\n"
            files = {"file": ("sites.csv", csv_string, "text/csv")}

            self.post(
                URLS.add_bulk_external_sites,
                params={"action": "SAVE_AND_LAUNCH"},
                files=files,
                # The content type is set automatically by requests based on the file type
                content_type=None,
                accept="application/json",
            )



import datetime
import ipaddress

from pydantic_xml import BaseXmlModel, attr, element, wrapped


class ResponseWarning(BaseXmlModel):
    code: str = element(tag="CODE")
    text: str = element(tag="TEXT")
    url: str = element(tag="URL")


class QdsFactor(BaseXmlModel):
    name: str = attr(name="name")
    value: str | None


class Qds(BaseXmlModel):
    severity: str = attr(name="severity")
    score: int | None


class Detection(BaseXmlModel):
    unique_vuln_id: int = element(tag="UNIQUE_VULN_ID")
    qid: int = element(tag="QID")
    type: str = element(tag="TYPE")
    severity: int | None = element(tag="SEVERITY", default=None)
    port: int | None = element(tag="PORT", default=None)
    protocol: str | None = element(tag="PROTOCOL", default=None)
    fqdn: str | None = element(tag="FQDN", default=None)
    ssl: bool | None = element(tag="SSL", default=None)
    instance: str | None = element(tag="INSTANCE", default=None)
    results: str | None = element(tag="RESULTS", default=None)
    status: str | None = element(tag="STATUS", default=None)
    first_found_datetime: datetime.datetime | None = element(
        tag="FIRST_FOUND_DATETIME", default=None
    )
    last_found_datetime: datetime.datetime | None = element(
        tag="LAST_FOUND_DATETIME", default=None
    )
    qds: Qds | None = element(tag="QDS", default=None)
    qds_factors: list[QdsFactor] = wrapped(
        "QDS_FACTORS", element(tag="QDS_FACTOR", default_factory=list)
    )
    times_found: int | None = element(tag="TIMES_FOUND", default=None)
    last_test_datetime: datetime.datetime | None = element(
        tag="LAST_TEST_DATETIME", default=None
    )
    last_update_datetime: datetime.datetime | None = element(
        tag="LAST_UPDATE_DATETIME", default=None
    )
    last_fixed_datetime: datetime.datetime | None = element(
        tag="LAST_FIXED_DATETIME", default=None
    )
    first_reopened_datetime: datetime.datetime | None = element(
        tag="FIRST_REOPENED_DATETIME", default=None
    )
    last_reopened_datetime: datetime.datetime | None = element(
        tag="LAST_REOPENED_DATETIME", default=None
    )
    times_reopened: int | None = element(tag="TIMES_REOPENED", default=None)
    service: str | None = element(tag="SERVICE", default=None)
    is_ignored: bool | None = element(tag="IS_IGNORED", default=None)
    is_disabled: bool | None = element(tag="IS_DISABLED", default=None)
    affect_running_kernel: bool | None = element(
        tag="AFFECT_RUNNING_KERNEL", default=None
    )
    affect_running_service: bool | None = element(
        tag="AFFECT_RUNNING_SERVICE", default=None
    )
    affect_exploitable_config: bool | None = element(
        tag="AFFECT_EXPLOITABLE_CONFIG", default=None
    )
    last_processed_datetime: datetime.datetime | None = element(
        tag="LAST_PROCESSED_DATETIME", default=None
    )
    asset_cve: str | None = element(tag="ASSET_CVE", default=None)


class CloudTag(BaseXmlModel):
    name: str = element(tag="NAME")
    value: str = element(tag="VALUE")
    last_success_date: datetime.datetime = element(tag="LAST_SUCCESS_DATE")


class Attribute(BaseXmlModel):
    name: str = element(tag="NAME")
    last_status: str = element(tag="LAST_STATUS")
    value: str = element(tag="VALUE")
    last_success_date: datetime.datetime | None = element(
        tag="LAST_SUCCESS_DATE", default=None
    )
    last_error_date: datetime.datetime | None = element(
        tag="LAST_ERROR_DATE", default=None
    )
    last_error: str | None = element(tag="LAST_ERROR", default=None)


class Metadata(BaseXmlModel):
    ec2: list[Attribute] = wrapped(
        "EC2", element(tag="ATTRIBUTE", default_factory=list)
    )
    google: list[Attribute] = wrapped(
        "GOOGLE", element(tag="ATTRIBUTE", default_factory=list)
    )
    azure: list[Attribute] = wrapped(
        "AZURE", element(tag="ATTRIBUTE", default_factory=list)
    )


class Tag(BaseXmlModel):
    tag_id: int | None = element(tag="TAG_ID")
    name: str = element(tag="NAME")
    color: str | None = element(tag="COLOR", default=None)
    background_color: str | None = element(tag="BACKGROUND_COLOR", default=None)


class DnsData(BaseXmlModel):
    hostname: str | None = element(tag="HOSTNAME", default=None)
    domain: str | None = element(tag="DOMAIN", default=None)
    fqdn: str | None = element(tag="FQDN", default=None)


class Host(BaseXmlModel):
    id: int = element(tag="ID")
    asset_id: int | None = element(tag="ASSET_ID", default=None)
    ip: ipaddress.IPv4Address = element(tag="IP", default=None)
    ipv6: ipaddress.IPv6Address | None = element(tag="IPV6", default=None)
    tracking_method: str | None = element(tag="TRACKING_METHOD", default=None)
    network_id: int | None = element(tag="NETWORK_ID", default=None)
    os: str | None = element(tag="OS", default=None)
    os_cpe: str | None = element(tag="OS_CPE", default=None)
    dns: str | None = element(tag="DNS", default=None)
    dns_data: DnsData | None = element(tag="DNS_DATA", default=None)
    cloud_provider: str | None = element(tag="CLOUD_PROVIDER", default=None)
    cloud_service: str | None = element(tag="CLOUD_SERVICE", default=None)
    cloud_resource_id: str | None = element(tag="CLOUD_RESOURCE_ID", default=None)
    ec2_instance_id: str | None = element(tag="EC2_INSTANCE_ID", default=None)
    netbios: str | None = element(tag="NETBIOS", default=None)
    qg_hostid: str | None = element(tag="QG_HOSTID", default=None)
    last_scan_datetime: datetime.datetime | None = element(
        tag="LAST_SCAN_DATETIME", default=None
    )
    last_vm_scanned_date: datetime.datetime | None = element(
        tag="LAST_VM_SCANNED_DATE", default=None
    )
    last_vm_scanned_duration: int | None = element(
        tag="LAST_VM_SCANNED_DURATION", default=None
    )
    last_vm_auth_scanned_date: datetime.datetime | None = element(
        tag="LAST_VM_AUTH_SCANNED_DATE", default=None
    )
    last_vm_auth_scanned_duration: int | None = element(
        tag="LAST_VM_AUTH_SCANNED_DURATION", default=None
    )
    last_pc_scanned_date: datetime.datetime | None = element(
        tag="LAST_PC_SCANNED_DATE", default=None
    )
    tags: list[Tag] = wrapped("TAGS", element(tag="TAG", default_factory=list))
    metadata: Metadata | None = element(tag="METADATA", alias="metadata_", default=None)
    cloud_provider_tags: list[CloudTag] = wrapped(
        "CLOUD_PROVIDER_TAGS", element(tag="CLOUD_TAG", default_factory=list)
    )
    detections: list[Detection] = wrapped(
        "DETECTION_LIST", element(tag="DETECTION", default_factory=list)
    )


class Response(BaseXmlModel):
    response_datetime: datetime.datetime = element(tag="DATETIME")
    host_list: list[Host] | None = wrapped(
        "HOST_LIST", element(tag="HOST", default=None)
    )
    warning: ResponseWarning | None = element(tag="WARNING", default=None)


class Param(BaseXmlModel):
    key: str = element(tag="KEY")
    value: str = element(tag="VALUE")


class Request(BaseXmlModel):
    request_datetime: datetime.datetime = element(tag="DATETIME")
    user_login: str = element(tag="USER_LOGIN")
    resource: str = element(tag="RESOURCE")
    param_list: list[Param] | None = element(tag="PARAM_LIST")
    post_data: str | None = element(tag="POST_DATA")


class HostListVMDetectionOutput(BaseXmlModel, tag="HOST_LIST_VM_DETECTION_OUTPUT"):
    request: Request | None = element(tag="REQUEST", default=None)
    response: Response = element(tag="RESPONSE", default=None)
